#include <stdlib.h>
#include "string.h"
#include "anamod.h"

#define NALLOC 70

struct IntArray_ {
  char *name; int alloc,fill; PetscBool *has; int *data;
};

#undef __FUNCT__
#define __FUNCT__ "CreateIntArray"
PetscErrorCode CreateIntArray(const char *name,int size,IntArray *array)
{
  IntArray inew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct IntArray_),&inew); CHKERRQ(ierr);
  ierr = PetscMemzero(inew,sizeof(struct IntArray_)); CHKERRQ(ierr);
  if (name) inew->name = strdup(name);
  inew->fill = 0; inew->alloc = NALLOC;
  ierr = PetscMalloc(inew->alloc*sizeof(PetscBool),&(inew->has)); CHKERRQ(ierr);
  ierr = PetscMemzero(inew->has,inew->alloc*sizeof(PetscBool)); CHKERRQ(ierr);
  ierr = PetscMalloc(inew->alloc*sizeof(int),&(inew->data)); CHKERRQ(ierr);
  ierr = PetscMemzero(inew->data,inew->alloc*sizeof(int)); CHKERRQ(ierr);
  *array = inew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeleteIntArray"
PetscErrorCode DeleteIntArray(IntArray array)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscFree(array->data); CHKERRQ(ierr);
  ierr = PetscFree(array->has); CHKERRQ(ierr);
  if (array->name) free(array->name);
  ierr = PetscFree(array); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "IntArrayAdd"
/*! Add a new value and return the index. Right now we do not yet dynamically
  reallocate the array if there is no more space.
*/
PetscErrorCode IntArrayAdd(IntArray array,int val,int *idx)
{
  int ins;
  PetscFunctionBegin;
  if (array->fill>=array->alloc-1) SETERRQ(MPI_COMM_WORLD,1,"No more space");
  ins = array->fill++;
  array->data[ins] = val;
  array->has[ins] = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "IntArraySetAt"
/*! Set a value at a given index. Right now we do not yet dynamically
  reallocate the array if the index is out of bound.
*/
PetscErrorCode IntArraySetAt(IntArray array,int idx,int val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  array->data[idx] = val; array->has[idx] = PETSC_TRUE;
  if (idx>array->fill) array->fill = idx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "IntArrayTryGetAt"
/*! Retrieve data from a given index. 

  Arguments:
  - \c array : the IntArray object
  - \c idx : the index; it is an error to ask out of bounds, but
  not to ask beyond the highest position filled; in that case failure will
  be reported
  - \c val (output) : the value retrieved. This argument is allowed to be null.
  - \c has (output) : true if a value was stored at this index.
  This argument is allowed to be null.
*/
PetscErrorCode IntArrayTryGetAt(IntArray array,int idx,int *val,PetscBool *has)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (has) *has = array->has[idx];
  if (array->has[idx] && val) *val = array->data[idx];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "IntArrayGetAt"
/*! As IntArrayTryGetAt(), except that it is an error to ask
  for an index where no value has been set.
*/
PetscErrorCode IntArrayGetAt(IntArray array,int idx,int *val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (!array->has[idx]) SETERRQ1(MPI_COMM_WORLD,1,"Asking for non-set element: %d",idx);
  *val = array->data[idx];
  PetscFunctionReturn(0);
}

/*
 * String arrays
 */
struct StringArray_ {
  char *name; int alloc,fill; PetscBool *has; char **data;
};

#undef __FUNCT__
#define __FUNCT__ "CreateStringArray"
PetscErrorCode CreateStringArray(const char *name,int size,StringArray *array)
{
  StringArray snew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct StringArray_),&snew); CHKERRQ(ierr);
  ierr = PetscMemzero(snew,sizeof(struct StringArray_)); CHKERRQ(ierr);
  if (name) snew->name = strdup(name);
  snew->fill = 0; snew->alloc = NALLOC;
  ierr = PetscMalloc(snew->alloc*sizeof(PetscBool),&(snew->has)); CHKERRQ(ierr);
  ierr = PetscMemzero(snew->has,snew->alloc*sizeof(PetscBool)); CHKERRQ(ierr);
  ierr = PetscMalloc(snew->alloc*sizeof(char*),&(snew->data)); CHKERRQ(ierr);
  ierr = PetscMemzero(snew->data,snew->alloc*sizeof(char*)); CHKERRQ(ierr);
  *array = snew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeleteStringArray"
PetscErrorCode DeleteStringArray(StringArray array)
{
  int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  for (i=0; i<array->fill; i++)
    if (array->has[i]) free(array->data[i]);
  ierr = PetscFree(array->data); CHKERRQ(ierr);
  ierr = PetscFree(array->has); CHKERRQ(ierr);
  if (array->name) free(array->name);
  ierr = PetscFree(array); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StringArrayAdd"
/*! Add a new value and return the index. Right now we do not yet dynamically
  reallocate the array if there is no more space.
*/
PetscErrorCode StringArrayAdd(StringArray array,const char *val,int *idx)
{
  int ins;
  PetscFunctionBegin;
  if (array->fill>=array->alloc-1) SETERRQ(MPI_COMM_WORLD,1,"No more space");
  ins = array->fill++;
  array->data[ins] = strdup(val);
  array->has[ins] = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StringArrayGetFill"
PetscErrorCode StringArrayGetFill(StringArray array,int *idx)
{
  PetscFunctionBegin;
  *idx = array->fill;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StringArraySetAt"
/*! Set a value at a given index. Right now we do not yet dynamically
  reallocate the array if the index is out of bound.
*/
PetscErrorCode StringArraySetAt(StringArray array,int idx,const char *val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  array->data[idx] = strdup(val); array->has[idx] = PETSC_TRUE;
  if (idx>array->fill) array->fill = idx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StringArrayTryGetAt"
/*! Retrieve data from a given index. 

  Arguments:
  - \c array : the StringArray object
  - \c idx : the index; it is an error to ask out of bounds, but
  not to ask beyond the highest position filled; in that case failure will
  be reported
  - \c val (output) : the value retrieved. This argument is allowed to be null.
  - \c has (output) : true if a value was stored at this index.
  This argument is allowed to be null.
*/
PetscErrorCode StringArrayTryGetAt(StringArray array,int idx,char **val,PetscBool *has)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (has) *has = array->has[idx];
  if (array->has[idx] && val) *val = array->data[idx];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StringArrayGetAt"
/*! As StringArrayTryGetAt(), except that it is an error to ask
  for an index where no value has been set.
*/
PetscErrorCode StringArrayGetAt(StringArray array,int idx,char **val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (!array->has[idx]) SETERRQ1(MPI_COMM_WORLD,1,"Asking for non-set element: %d",idx);
  *val = array->data[idx];
  PetscFunctionReturn(0);
}

/*
 * AnalysisItem arrays
 */
struct AnalysisItemArray_ {
  char *name; int alloc,fill,*has; AnalysisItem *data;
};

#undef __FUNCT__
#define __FUNCT__ "CreateAnalysisItemArray"
PetscErrorCode CreateAnalysisItemArray
(const char *name,int size,AnalysisItemArray *array)
{
  AnalysisItemArray anew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct AnalysisItemArray_),&anew); CHKERRQ(ierr);
  ierr = PetscMemzero(anew,sizeof(struct AnalysisItemArray_)); CHKERRQ(ierr);
  if (name) anew->name = strdup(name);
  anew->fill = 0; anew->alloc = NALLOC;
  ierr = PetscMalloc(anew->alloc*sizeof(PetscBool),&(anew->has)); CHKERRQ(ierr);
  ierr = PetscMemzero(anew->has,anew->alloc*sizeof(PetscBool)); CHKERRQ(ierr);
  ierr = PetscMalloc(anew->alloc*sizeof(AnalysisItem),&(anew->data)); CHKERRQ(ierr);
  ierr = PetscMemzero(anew->data,anew->alloc*sizeof(AnalysisItem)); CHKERRQ(ierr);
  *array = anew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeleteAnalysisItemArray"
PetscErrorCode DeleteAnalysisItemArray(AnalysisItemArray array)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscFree(array->data); CHKERRQ(ierr);
  ierr = PetscFree(array->has); CHKERRQ(ierr);
  if (array->name) free(array->name);
  ierr = PetscFree(array); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisItemArrayAdd"
/*! Add a new value and return the index. Right now we do not yet dynamically
  reallocate the array if there is no more space.
*/
PetscErrorCode AnalysisItemArrayAdd(AnalysisItemArray array,AnalysisItem val,int *idx)
{
  int ins;
  PetscFunctionBegin;
  if (array->fill>=array->alloc-1) SETERRQ(MPI_COMM_WORLD,1,"No more space");
  ins = array->fill++;
  array->data[ins] = val;
  array->has[ins] = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisItemArraySetAt"
/*! Set a value at a given index. Right now we do not yet dynamically
  reallocate the array if the index is out of bound.
*/
PetscErrorCode AnalysisItemArraySetAt(AnalysisItemArray array,int idx,AnalysisItem val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  array->data[idx] = val; array->has[idx] = PETSC_TRUE;
  if (idx>array->fill) array->fill = idx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisItemArrayTryGetAt"
/*! Retrieve data from a given index. 

  Arguments:
  - \c array : the AnalysisItemArray object
  - \c idx : the index; it is an error to ask out of bounds, but
  not to ask beyond the highest position filled; in that case failure will
  be reported
  - \c val (output) : the value retrieved. This argument is allowed to be null.
  - \c has (output) : true if a value was stored at this index.
  This argument is allowed to be null.
*/
PetscErrorCode AnalysisItemArrayTryGetAt(AnalysisItemArray array,int idx,AnalysisItem *val,PetscBool *has)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (has) *has = (PetscBool)array->has[idx];
  if (array->has[idx] && val) *val = array->data[idx];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisItemArrayGetAt"
/*! As AnalysisItemArrayTryGetAt(), except that it is an error to ask
  for an index where no value has been set.
*/
PetscErrorCode AnalysisItemArrayGetAt(AnalysisItemArray array,int idx,AnalysisItem *val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (!array->has[idx]) SETERRQ1(MPI_COMM_WORLD,1,"Asking for non-set element: %d",idx);
  *val = array->data[idx];
  PetscFunctionReturn(0);
}

/*
 * AnalysisDataType arrays
 */
struct AnalysisDataTypeArray_ {
  char *name; int alloc,fill,*has; AnalysisDataType *data;
};

#undef __FUNCT__
#define __FUNCT__ "CreateAnalysisDataTypeArray"
PetscErrorCode CreateAnalysisDataTypeArray
(const char *name,int size,AnalysisDataTypeArray *array)
{
  AnalysisDataTypeArray anew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct AnalysisDataTypeArray_),&anew); CHKERRQ(ierr);
  ierr = PetscMemzero(anew,sizeof(struct AnalysisDataTypeArray_)); CHKERRQ(ierr);
  if (name) anew->name = strdup(name);
  anew->fill = 0; anew->alloc = NALLOC;
  ierr = PetscMalloc(anew->alloc*sizeof(PetscBool),&(anew->has)); CHKERRQ(ierr);
  ierr = PetscMemzero(anew->has,anew->alloc*sizeof(PetscBool)); CHKERRQ(ierr);
  ierr = PetscMalloc(anew->alloc*sizeof(AnalysisDataType),&(anew->data)); CHKERRQ(ierr);
  ierr = PetscMemzero(anew->data,anew->alloc*sizeof(AnalysisDataType)); CHKERRQ(ierr);
  *array = anew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeleteAnalysisDataTypeArray"
PetscErrorCode DeleteAnalysisDataTypeArray(AnalysisDataTypeArray array)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscFree(array->data); CHKERRQ(ierr);
  ierr = PetscFree(array->has); CHKERRQ(ierr);
  if (array->name) free(array->name);
  ierr = PetscFree(array); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisDataTypeArrayAdd"
/*! Add a new value and return the index. Right now we do not yet dynamically
  reallocate the array if there is no more space.
*/
PetscErrorCode AnalysisDataTypeArrayAdd(AnalysisDataTypeArray array,AnalysisDataType val,int *idx)
{
  int ins;
  PetscFunctionBegin;
  if (array->fill>=array->alloc-1) SETERRQ(MPI_COMM_WORLD,1,"No more space");
  ins = array->fill++;
  array->data[ins] = val;
  array->has[ins] = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisDataTypeArraySetAt"
/*! Set a value at a given index. Right now we do not yet dynamically
  reallocate the array if the index is out of bound.
*/
PetscErrorCode AnalysisDataTypeArraySetAt(AnalysisDataTypeArray array,int idx,AnalysisDataType val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  array->data[idx] = val; array->has[idx] = PETSC_TRUE;
  if (idx>array->fill) array->fill = idx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisDataTypeArrayTryGetAt"
/*! Retrieve data from a given index. 

  Arguments:
  - \c array : the AnalysisDataTypeArray object
  - \c idx : the index; it is an error to ask out of bounds, but
  not to ask beyond the highest position filled; in that case failure will
  be reported
  - \c val (output) : the value retrieved. This argument is allowed to be null.
  - \c has (output) : true if a value was stored at this index.
  This argument is allowed to be null.
*/
PetscErrorCode AnalysisDataTypeArrayTryGetAt(AnalysisDataTypeArray array,int idx,AnalysisDataType *val,PetscBool *has)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (has) *has = (PetscBool)array->has[idx];
  if (array->has[idx] && val) *val = array->data[idx];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnalysisDataTypeArrayGetAt"
/*! As AnalysisDataTypeArrayTryGetAt(), except that it is an error to ask
  for an index where no value has been set.
*/
PetscErrorCode AnalysisDataTypeArrayGetAt(AnalysisDataTypeArray array,int idx,AnalysisDataType *val)
{
  PetscFunctionBegin;
  if (idx>=array->alloc) SETERRQ1(MPI_COMM_WORLD,1,"Index out of bounds: %d",idx);
  if (!array->has[idx]) SETERRQ1(MPI_COMM_WORLD,1,"Asking for non-set element: %d",idx);
  *val = array->data[idx];
  PetscFunctionReturn(0);
}

