/*! \file options.c \ingroup implementation

  \brief Commandline options handling

*/
/*!  \page options Commandline options

  The runtime behaviour of AnaMod can be influenced by commandline options,
  specified at the start of the calling application.

  All options start with \c "-anamod" and they can have a sequence of
  comma-separated values separated from the option by a 
  blank: \c "-anamod_option value1,value2". Note the single dash, which is Petsc
  style, as opposed to the double dash of GNU style.

  - \c -anamod_force : Certain operations can be very expensive, 
  for instance because
  they need to be performed on a single processor. AnaMod will be default 
  refuse to compute these quantities, but their computation can be forced
  by \c "-anamod_force <option>" where \c option is
    - \c "expensive" : certain modules are very expensive to compute in
    certain circumstances, so AnaMod will normally refuse to compute them.
    (See \ref normal.) This option forces their computation, no matter how
    much time they may take.
    - \c "sequential" : certain modules are only implemented as single
    processor code, so they will normally not be computed in a parallel
    run. This option will force processor zero to gather the full matrix,
    and perform the computation locally. I hope I do not have to point out
    the potential pitfalls of this option. Sequential modules are the norms
    of the symmetric/anti-symmetric part in the \ref simple category.
  - Each module can have options \c "-anamod_<module name> <option>".
  If the module has declared a function to handle such options, that
  function is called here. This implies that the AnaModOptionsHandling()
  function needs to be called after all modules have been declared.
  - \c "-anamod_use_only <mod>" : prevent other categories from
  being computed
  - \c -anamod_compute : (not implemented yet)

 */

#include <stdlib.h>
#include <string.h>
#include "anamod.h"
#include "petsc.h"

static PetscBool single_proc = PETSC_FALSE, expensive = PETSC_FALSE;

#undef __FUNCT__
#define __FUNCT__ "AnaModOptionsHandling"
/*!
  Process any commandline options that were given for AnaMod.
  These use the regular Petsc commandline options database.

  This routine handles general options and module-specific options,
  so it has to be called after all modules have been declared.

  See DeclareCategoryOptionFunction(),CategoryGetOptionFunction(),
  AnaModOptionsHandling() and section \ref options.
*/
PetscErrorCode AnaModOptionsHandling()
{
  PetscBool flg; PetscErrorCode ierr;
  char *value;
  PetscFunctionBegin;

#define VALUELEN 150
  ierr = PetscMalloc(VALUELEN*sizeof(char),&value); CHKERRQ(ierr);

  {
    ierr = PetscOptionsHasName(PETSC_NULL,"-anamod_force",&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = PetscOptionsGetString
	(PETSC_NULL,"-anamod_force",value,VALUELEN,&flg); CHKERRQ(ierr);
      if (flg) {
	char *c;
	c = strtok(value,",");
	while (c) {
	  ierr = PetscStrcmp(c,"sequential",&flg); CHKERRQ(ierr);
	  if (flg) single_proc = PETSC_TRUE;
	  ierr = PetscStrcmp(c,"expensive",&flg); CHKERRQ(ierr);
	  if (flg) expensive = PETSC_TRUE;
	  c = strtok(NULL,",");
	}
      }
    }
  }

  {
    const char **categories; char *option; int icat,ncats;
    ierr = PetscMalloc(30*sizeof(char),&option); CHKERRQ(ierr);
    ierr = GetCategories(&ncats,&categories); CHKERRQ(ierr);
    for (icat=0; icat<ncats; icat++) {
      ierr = PetscMemzero(value,VALUELEN*sizeof(char)); CHKERRQ(ierr);      
      ierr = PetscMemzero(option,30*sizeof(char)); CHKERRQ(ierr);
      sprintf(option,"-anamod_%s",categories[icat]);
      ierr = PetscOptionsGetString
	(PETSC_NULL,option,value,VALUELEN,&flg); CHKERRQ(ierr);
      if (flg) {
	PetscErrorCode (*f)(char*);
	ierr = CategoryGetOptionFunction(categories[icat],&f); CHKERRQ(ierr);
	if (f) {
	  ierr = (*f)(value); CHKERRQ(ierr);
	}
      }
    }
    ierr = PetscFree(option); CHKERRQ(ierr);
  }

  {
    ierr = PetscOptionsGetString
      (PETSC_NULL,"-anamod_use_only",value,VALUELEN,&flg); CHKERRQ(ierr);
    if (flg) {
      int ncats,icat; const char **cats;
      ierr = GetCategories(&ncats,&cats); CHKERRQ(ierr);
      for (icat=0; icat<ncats; icat++) {
	if (strcmp(value,cats[icat])) {
	  ierr = CategoryEnableByName
	    (cats[icat],CATCMP_SKIP_FROM_LOOPS); CHKERRQ(ierr);
	}
      }
    }
  }

  ierr = PetscFree(value); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModShowOptions"
/*! Display all available options. This depends on the installed modules,
  so you need to do the various register calls first. See \ref usage.

  For options see \ref options.
*/
PetscErrorCode AnaModShowOptions(MPI_Comm comm)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  PetscPrintf(comm,"Available AnaMod options\n  Global:\n");
  PetscPrintf(comm,"  -anamod_use_only mod : use only the specified module\n");
  PetscPrintf(comm,"  -anamod_force sequential : do sequential computations\n  even in parallel runs\n");
  PetscPrintf(comm,"  -anamod_force expensive : do certain calculations no matter how expensive\n");
  PetscPrintf(comm,"\n  Categories and modules:\n");
  {
    const char **categories; char *option; int icat,ncats;
    ierr = PetscMalloc(30*sizeof(char),&option); CHKERRQ(ierr);
    ierr = GetCategories(&ncats,&categories); CHKERRQ(ierr);
    for (icat=0; icat<ncats; icat++) {
      ierr = PetscMemzero(option,30*sizeof(char)); CHKERRQ(ierr);
      PetscPrintf
	(comm,"  -anamod_%s : options for category %s\n",
	 categories[icat],categories[icat]);
    }
    ierr = PetscFree(option); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModHasForcedSequentialComputation"
/*!
  Query whether parallel modules should be done on processor zero.
*/
PetscErrorCode AnaModHasForcedSequentialComputation(PetscBool *flg)
{
  PetscFunctionBegin;
  *flg = single_proc;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModGetSequentialMatrix"
/*!
  Collect the matrix on processor zero.

  There is an implicit assumption here that processor zero is the one
  that will do the actual work.

  Argument:
  - \c A : input matrix
  - \c Awork : pointer to the sequential matrix, this can be a pointer
  to the original matrix if it was already sequential; it is NULL
  if no forced sequential computation is asked (see \ref options)
  - \c mem : true if \c Awork is newly allocated
  - \c local : true if this processor needs to do the work
  - \c global : true if any processor does work; this condition is false
  in the case of a distributed matrix and no forced sequential operation

*/
PetscErrorCode AnaModGetSequentialMatrix
(Mat A,Mat *Awork,PetscBool *mem,PetscBool *local,PetscBool *global)
{
  MPI_Comm comm; const MatType type; int mytid;
  PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MPI_Comm_rank(comm,&mytid);
  ierr = MatGetType(A,&type); CHKERRQ(ierr);
  ierr = PetscStrcmp(type,MATSEQAIJ,&flg); CHKERRQ(ierr);
  if (flg) {
    *Awork = A; *mem = PETSC_FALSE;
    *local = PETSC_TRUE; *global = PETSC_TRUE;
  } else {
    ierr = AnaModHasForcedSequentialComputation(&flg); CHKERRQ(ierr);
    if (!flg) {
      *Awork = NULL;
      *mem = PETSC_FALSE; *local = PETSC_FALSE; *global = PETSC_FALSE;
      PetscFunctionReturn(0);
    }
    *global = PETSC_TRUE;
    {
      Mat *Arow; IS is[1]; int N,M;
      ierr = MatGetSize(A,&M,&N); CHKERRQ(ierr);
      if (mytid==0) {
	ierr = ISCreateStride(MPI_COMM_SELF,M,0,1,is); CHKERRQ(ierr);
	*local = PETSC_TRUE;
      } else {
	ierr = ISCreateStride(MPI_COMM_SELF,0,0,1,is); CHKERRQ(ierr);
	*local = PETSC_FALSE;
      }
      ierr = MatGetSubMatrices
	(A,1,is,is,MAT_INITIAL_MATRIX,&Arow); CHKERRQ(ierr);
      ierr = ISDestroy(&(is[0])); CHKERRQ(ierr);
      *Awork = Arow[0]; *mem = PETSC_TRUE;
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModHasForcedExpensiveComputation"
/*!
  Query whether certain expensive operations should be done 
  regardless the cost.
*/
PetscErrorCode AnaModHasForcedExpensiveComputation(PetscBool *flg)
{
  PetscFunctionBegin;
  *flg = expensive;
  PetscFunctionReturn(0);
}
