/*! \file structure.c \ingroup categories 
  \brief Structural properties of a matrix

  This file defines the routines for computing structural properties 
  of a matrix, as well as the routine RegisterStructureModules() that 
  registers them.

  \section structure Structural properties

  This category computes matrix properties that are only a function of
  the nonzero structure. Thus, these properties will likely stay invariant
  during a nonlinear solve process, or while time-stepping a system
  of equations.

  \section Usage

  Activate this module with

  PetscErrorCode RegisterStructureModules();

  Compute these elements with

  ComputeQuantity("structure",\<element\>,A,(void*)&res,&flg);

  Available elements are:
  - "nrows" : number of rows in the matrix
  - "nnzeros" : number of nonzero elements in the matrix
  - "max-nnzeros-per-row" : maximum number of nonzeros per row
  - "min-nnzeros-per-row" : minimum number of nonzeros per row
  - "left-bandwidth" : \f$\max_i\{i-j\colon a_{ij}\not=0\}\f$
  - "right-bandwidth" : \f$\max_i\{j-i\colon a_{ij}\not=0\}\f$
  - "n-dummy-rows" : number of rows with only one element
  - "dummy-rows-kind" : 0 if all dummy rows have a one on the diagonal,
    1 if the value is not one, 2 if not on the diagonal
  - "diag-zerostart" : \f$\min\{i\colon \forall_{j>i}a_{jj}=0\}\f$
  - "diag-definite" : 1 if diagonal positive, 0 otherwise
  - "block-size" : integer size of blocks that comprise matrix block structure,
    1 in the general case
  - "symmetry" : 1 for symmetric matrix, 0 otherwise
  - "n-struct-unsymm" : number of structurally unsymmetric elements
    (this is actually calculated in the simple.c)

*/
#include <stdlib.h>
#include "anamod.h"
#include "anamodutils.h"
#include "anamodsalsamodules.h"
#include "petscerror.h"
#include "petscmat.h"

#undef __FUNCT__
#define __FUNCT__ "nRows"
static PetscErrorCode nRows
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","nrows",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    int idum;
    ierr = MatGetSize(A,&v,&idum); CHKERRQ(ierr);
    *flg = PETSC_TRUE;
    ierr = PetscObjectComposedDataSetInt
      ((PetscObject)A,id,v); CHKERRQ(ierr);
      rv->i = v;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Symmetry"
static PetscErrorCode Symmetry
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","symmetry",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    PetscBool f; const MatType type;
    ierr = MatGetType(A,&type); CHKERRQ(ierr);
    ierr = PetscStrcmp(type,MATSEQAIJ,flg); CHKERRQ(ierr);
    if (*flg) {
      ierr = MatIsSymmetric(A,0.0,&f); CHKERRQ(ierr);
      v = f;
      ierr = PetscObjectComposedDataSetInt
	((PetscObject)A,id,v); CHKERRQ(ierr);
      rv->i = v;
    }
  }
  
  PetscFunctionReturn(0);
}

/*
 * Nonzero structure, bandwidth
 */
#undef __FUNCT__
#define __FUNCT__ "compute_nnz_structure"
static PetscErrorCode compute_nnz_structure(AnaModNumericalProblem prob)
{
  Mat A = (Mat)prob;
  int nnz,p,q,minnz,maxnz,minnnz,maxnnz;
  MPI_Comm comm;
  int N,first,last,local_size,row,nz,*Bl,*Br,bl,br, id; 
  PetscBool has; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MatGetSize(A,&N,PETSC_NULL); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  local_size = last-first;
  ierr = PetscMalloc(local_size*sizeof(int),&Bl); CHKERRQ(ierr);
  ierr = PetscMalloc(local_size*sizeof(int),&Br); CHKERRQ(ierr);
  bl = 0; br = 0; nz = 0; minnz = maxnz = 0;
  for (row=first; row<last; row++) {
    int icol,ncols,localnz=0; const int *cols; const PetscScalar *val;
    ierr = MatGetRow(A,row,&ncols,&cols,&val); CHKERRQ(ierr);
    for (icol=0; icol<ncols; icol++) 
      if (val[icol]!=0.) localnz++;
    nz += localnz;
    if (row==first) minnz = localnz;
    if (localnz<minnz) minnz = localnz;
    if (localnz>maxnz) maxnz = localnz;
    if (ncols) {
      Bl[row-first] = bl = PetscMax(bl,row-cols[0]);
      Br[row-first] = br = PetscMax(br,cols[ncols-1]-row);
    }
    ierr = MatRestoreRow(A,row,&ncols,&cols,&val); CHKERRQ(ierr);
  }

  ierr = MPI_Allreduce((void*)&nz,(void*)&nnz,1,MPI_INT,MPI_SUM,comm);
  ierr = GetDataID("structure","nnzeros",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,nnz); CHKERRQ(ierr);

  ierr = MPI_Allreduce((void*)&maxnz,(void*)&maxnnz,1,MPI_INT,MPI_MAX,comm);
  ierr = GetDataID("structure","max-nnzeros-per-row",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,maxnnz); CHKERRQ(ierr);

  ierr = MPI_Allreduce((void*)&minnz,(void*)&minnnz,1,MPI_INT,MPI_MIN,comm);
  ierr = GetDataID("structure","min-nnzeros-per-row",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,minnnz); CHKERRQ(ierr);

  ierr = MPI_Allreduce((void*)&bl,(void*)&p,1,MPI_INT,MPI_MAX,comm);
  ierr = GetDataID("structure","left-bandwidth",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,p); CHKERRQ(ierr);

  ierr = MPI_Allreduce((void*)&br,(void*)&q,1,MPI_INT,MPI_MAX,comm);
  ierr = GetDataID("structure","right-bandwidth",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,q); CHKERRQ(ierr);

  ierr = PetscFree(Bl); CHKERRQ(ierr);
  ierr = PetscFree(Br); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "NNonZeros"
static PetscErrorCode NNonZeros
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","nnzeros",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "MaxNNonZerosPerRow"
static PetscErrorCode MaxNNonZerosPerRow
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","max-nnzeros-per-row",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "MinNNonZerosPerRow"
static PetscErrorCode MinNNonZerosPerRow
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","min-nnzeros-per-row",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LBandWidth"
static PetscErrorCode LBandWidth
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","left-bandwidth",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RBandWidth"
static PetscErrorCode RBandWidth
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","right-bandwidth",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LeftSkyline"
static PetscErrorCode LeftSkyline
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob; int N;
  int *v = NULL; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = MatGetSize(A,&N,PETSC_NULL); CHKERRQ(ierr);
  ierr = GetDataID("structure","left-skyline",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetIntstar
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) {
    rv->ii = v; *lv = N;
  } else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetIntstar
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) {
      rv->ii = v; *lv = N;
    }
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RightSkyline"
static PetscErrorCode RightSkyline
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob; int N;
  int *v = NULL; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = MatGetSize(A,&N,PETSC_NULL); CHKERRQ(ierr);
  ierr = GetDataID("structure","right-skyline",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetIntstar
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) {
    rv->ii = v; *lv = N;
  } else {
    ierr = compute_nnz_structure(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetIntstar
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) {
      rv->ii = v; *lv = N;
    }
  }
  
  PetscFunctionReturn(0);
}

/*
 * Dummy rows
 */
#undef __FUNCT__
#define __FUNCT__ "compute_dummy_rows"
static PetscErrorCode compute_dummy_rows(Mat A)
{
  int gkind;
  MPI_Comm comm;
  int first,last,row,ncols,ndummies,Ndummies,*dummies,*Dummies,kind,id; 
  PetscBool has; PetscErrorCode ierr;
  const int *cols; const PetscScalar *vals;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  ierr = PetscMalloc((last-first+1)*sizeof(int),&dummies); CHKERRQ(ierr);

  ndummies = 0; kind = 0;
  for (row=first; row<last; row++) {
    ierr = MatGetRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
    if (ncols==1) {
      dummies[ndummies++] = row;
      if (vals[0]!=1. && !kind ) kind = 1;
      if (cols[0]!=row) kind = 2;
    }
    ierr = MatRestoreRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
  }

  ierr = MPI_Allreduce
    ((void*)&kind,(void*)&gkind,1,MPI_INT,MPI_MAX,comm); CHKERRQ(ierr);
  ierr = GetDataID("structure","dummy-rows-kind",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,gkind); CHKERRQ(ierr);

  {
    int *sizes,*offsets,ntids,i,s;

    ierr = MPI_Comm_size(comm,&ntids);
    ierr = PetscMalloc(ntids*sizeof(int),&sizes); CHKERRQ(ierr);
    ierr = PetscMalloc(ntids*sizeof(int),&offsets); CHKERRQ(ierr);
    ierr = MPI_Allgather(&ndummies,1,MPI_INT,sizes,1,MPI_INT,comm);
    s=0; for (i=0; i<ntids; i++) {offsets[i]=s; s+=sizes[i];} Ndummies = s;
    if (Ndummies==0) 
      Dummies = NULL;
    else {
      ierr = PetscMalloc(Ndummies*sizeof(int),&Dummies); CHKERRQ(ierr);
      ierr = MPI_Allgatherv
	(dummies,ndummies,MPI_INT,Dummies,sizes,offsets,MPI_INT,
	 comm); CHKERRQ(ierr);
    }
    ierr = PetscFree(dummies); CHKERRQ(ierr);
    ierr = PetscFree(offsets); CHKERRQ(ierr);
    ierr = PetscFree(sizes); CHKERRQ(ierr);
  }

  ierr = GetDataID("structure","n-dummy-rows",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,Ndummies); CHKERRQ(ierr);

  ierr = GetDataID("structure","dummy-rows",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetIntstar
    ((PetscObject)A,id,Dummies); CHKERRQ(ierr);
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "NDummyRows"
static PetscErrorCode NDummyRows
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","n-dummy-rows",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_dummy_rows(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DummyRowsKind"
static PetscErrorCode DummyRowsKind
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","dummy-rows-kind",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_dummy_rows(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DummyRows"
static PetscErrorCode DummyRows
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int *v = NULL; PetscBool has; int id,id2; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","dummy-rows",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetIntstar
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  ierr = GetDataID("structure","n-dummy-rows",&id2,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id2,*lv,*flg); CHKERRQ(ierr);
  if (*flg) {
    rv->ii = v;
  } else {
    ierr = compute_dummy_rows(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetIntstar
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id2,*lv,*flg); CHKERRQ(ierr);
    if (*flg) rv->ii = v;
  }
  
  PetscFunctionReturn(0);
}

/*
 * Diagonal analysis
 */
#undef __FUNCT__
#define __FUNCT__ "compute_posdiag"
static PetscErrorCode compute_posdiag
(AnaModNumericalProblem prob)
{
  Mat A = (Mat)prob;
  MPI_Comm comm;
  Vec D;
  PetscScalar *d;
  int i,first,last,m,M,di,df,Di,Df,pd,id; 
  PetscBool has; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MatGetSize(A,&M,PETSC_NULL); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  m = last-first;
  ierr = VecCreateMPI(comm,m,PETSC_DECIDE,&D); CHKERRQ(ierr);
  ierr = MatGetDiagonal(A,D); CHKERRQ(ierr);
  ierr = VecGetArray(D,&d); CHKERRQ(ierr);
  di = 1; df = first;
  for (i=0; i<m; i++) {
    if (d[i].real()<=0) di = 0;
    if (d[i].real()!=0) df = i+1;
  }
  ierr = MPI_Allreduce((void*)&di,(void*)&Di,1,MPI_INT,MPI_PROD,comm);
  ierr = GetDataID("structure","diag-definite",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,Di); CHKERRQ(ierr);

  ierr = MPI_Allreduce((void*)&df,(void*)&Df,1,MPI_INT,MPI_MAX,comm);
  ierr = GetDataID("structure","diag-zerostart",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt
    ((PetscObject)A,id,Df); CHKERRQ(ierr);


  if (Df<M) pd = 2+Df; else pd = Di;

  ierr = VecRestoreArray(D,&d); CHKERRQ(ierr);
  ierr = VecDestroy(&D); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DiagZeroStart"
static PetscErrorCode DiagZeroStart
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","diag-zerostart",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_posdiag(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DiagDefinite"
static PetscErrorCode DiagDefinite
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","diag-definite",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = compute_posdiag(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

/*
 * Simple block analysis
 */
#undef __FUNCT__
#define __FUNCT__ "regularblocks"
static PetscErrorCode regularblocks
(AnaModNumericalProblem prob)
{
  Mat A = (Mat)prob;
  int first,last,local_size,row, bs,gbs,id,mytid;
  MPI_Comm comm; PetscBool has; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MPI_Comm_rank(comm,&mytid); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  local_size = last-first;
  if (mytid==0) {
    int row;
    bs = 1;
    for (row=first; row<last; row++) {
      const int *cols; const PetscScalar *vals; int ncols,icol=0;
      PetscBool found = PETSC_FALSE;
      ierr = MatGetRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
      while (icol<ncols && cols[icol]<row) icol++;
      if (cols[icol]>row+1 ||
	  (cols[icol]==row && icol+1<ncols && 
	   (cols[icol+1]>row+1 || vals[icol+1]==0.))) {
	bs = row+1; found = PETSC_TRUE;
      }
      ierr = MatRestoreRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
      if (found) break;
    }
  }
  ierr = MPI_Bcast((void*)&bs,1,MPI_INT,0,comm);
  for (row=first; row<last; row++) {
    const int *cols; const PetscScalar *vals; int ncols,icol=0;
      PetscBool found = PETSC_FALSE;
    if ((row+1)%bs!=0) continue;
    ierr = MatGetRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
    while (icol<ncols && cols[icol]<row) icol++;
    if ( (cols[icol]==row+1 && vals[icol]!=0.) ||
	 (cols[icol]==row && 
	  (icol+1<ncols && cols[icol+1]==row+1 && vals[icol+1]!=0.)) ) {
      bs = row+1; found = PETSC_TRUE;
    }
    ierr = MatRestoreRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
    if (found) break;
  }
  ierr = MPI_Allreduce((void*)&bs,(void*)&gbs,1,MPI_INT,MPI_MIN,comm);

  ierr = GetDataID("structure","blocksize",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetInt((PetscObject)A,id,gbs); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "BlockSize"
static PetscErrorCode BlockSize
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  int v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("structure","blocksize",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetInt
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (*flg) rv->i = v;
  else {
    ierr = regularblocks(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->i = v;
  }
  
  PetscFunctionReturn(0);
}

/*
 * Registration
 */
#undef __FUNCT__
#define __FUNCT__ "RegisterStructureModules"
PetscErrorCode RegisterStructureModules()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = RegisterModule
    ("structure","nrows",ANALYSISINTEGER,&nRows); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","symmetry",ANALYSISINTEGER,&Symmetry); CHKERRQ(ierr);

  ierr = RegisterModule
    ("structure","nnzeros",ANALYSISINTEGER,&NNonZeros); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","max-nnzeros-per-row",ANALYSISINTEGER,
     &MaxNNonZerosPerRow); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","min-nnzeros-per-row",ANALYSISINTEGER,
     &MinNNonZerosPerRow); CHKERRQ(ierr);

  ierr = RegisterModule
    ("structure","left-bandwidth",ANALYSISINTEGER,&LBandWidth); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","right-bandwidth",ANALYSISINTEGER,&RBandWidth); CHKERRQ(ierr);

  ierr = RegisterModule
    ("structure","left-skyline",ANALYSISINTARRAY,&LeftSkyline); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","right-skyline",ANALYSISINTARRAY,&RightSkyline); CHKERRQ(ierr);

  ierr = RegisterModule
    ("structure","n-dummy-rows",ANALYSISINTEGER,&NDummyRows); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","dummy-rows-kind",ANALYSISINTEGER,&DummyRowsKind); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","dummy-rows",ANALYSISINTARRAY,&DummyRows); CHKERRQ(ierr);

  ierr = RegisterModule
    ("structure","diag-zerostart",ANALYSISINTEGER,&DiagZeroStart); CHKERRQ(ierr);
  ierr = RegisterModule
    ("structure","diag-definite",ANALYSISINTEGER,&DiagDefinite); CHKERRQ(ierr);

  ierr = RegisterModule
    ("structure","blocksize",ANALYSISINTEGER,&BlockSize); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeRegisterStructureModules"
PetscErrorCode DeRegisterStructureModules(void)
{
  PetscFunctionBegin;
  PetscFunctionReturn(0);
}
