#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "petsc.h"
#include "anamod.h"

/*! \file tracing.c
  \brief Trace routines for the anamod package

  \section tracing Tracing the analysis modules

  The AnaMod package does not by default print out anything, other than
  severe error messages (Petsc macro SETERRQ) that accompany an abort.

  However, you can specify a trace function, which
  can further be tuned by specifying a trace context.

  See AnaModDeclareTraceFunction(), AnaModDeclareTraceContext(),
  AnaModTraceMessage(), AnaModSetTraceArrays(), AnaModTraceArrays().
*/

static PetscErrorCode(*anamodtrace)(void*,const char *fmt,va_list) = NULL;
static size_t anamodtracectx = (size_t)NULL;
static PetscBool anamodtracearrays = PETSC_FALSE;

#undef __FUNCT__
#define __FUNCT__ "AnaModDeclareTraceFunction"
/*! Specify a trace function.

  The trace function has a prototype 
\verbatim
  PetscErrorCode tracefunction(void*,char*,va_list)
\endverbatim
  which means that it has an arbitrary number of arguments, much like
  \c printf. The first argument is a context, which can be set by 
  AnaModDeclareTraceContext().

  Here is an example of how you would write a trace function:
\verbatim
#include <stdarg.h>
PetscErrorCode tracefunction(void *ctx,char *fmt,va_list argp)
{
  char *prefix = (char*)ctx;
  PetscFunctionBegin;
  printf("%s ",prefix);
  vprintf(fmt, argp);
  PetscFunctionReturn(0);
}
\endverbatim
  Consult \c string.h (probably in \c /usr/include) to see which "v"
  versions of \c printf are available.

  You can undeclare a trace function by passing NULL.
*/
PetscErrorCode AnaModDeclareTraceFunction(PetscErrorCode(*fn)(void*,const char*,va_list))
{
  PetscFunctionBegin;
  anamodtrace = fn;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModDeclareTraceContext"
/* Specify a trace context.
  
  See AnaModDeclareTraceFunction().
*/
PetscErrorCode AnaModDeclareTraceContext(void *ctx)
{
  PetscFunctionBegin;
  anamodtracectx = (size_t)ctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModSetTraceArrays"
/* Declare whether array typed modules should be traced. By default false.
  
  See AnaModDeclareTraceFunction(), AnaModTraceArrays().
*/
PetscErrorCode AnaModSetTraceArrays(PetscBool f)
{
  PetscFunctionBegin;
  anamodtracearrays = f;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModTraceArrays"
/* Test whether array typed modules should be traced. By default false.
  
  See AnaModSetTraceArrays().
*/
PetscErrorCode AnaModTraceArrays(PetscBool *f)
{
  PetscFunctionBegin;
  *f = anamodtracearrays;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModTraceMessage"
/*! This function prints a trace message if a trace function has been
  declared; see AnaModDeclareTraceFunction().
*/
PetscErrorCode AnaModTraceMessage(const char *fmt,...)
{
  va_list argp; PetscErrorCode ierr;
  PetscFunctionBegin;
  if (anamodtrace) {
    va_start(argp, fmt);
    ierr = (*anamodtrace)((void*)anamodtracectx,(char*)fmt,argp); CHKERRQ(ierr);
    va_end(argp);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModHasTrace"
/*! Test whether a trace function has been declared; 
  see AnaModDeclareTraceFunction(). Normally you would use AnaModTraceMessage()
  which performs this test internally,
  but this function can be useful if a large amount of processing
  has to be performed to construct the trace message to begin with.
*/
PetscErrorCode AnaModHasTrace(PetscBool *flg)
{
  PetscFunctionBegin;
  if (anamodtrace)
    *flg = PETSC_TRUE;
  else
    *flg = PETSC_FALSE;
  PetscFunctionReturn(0);
}

