#include <stdlib.h>
#include "memory.h"
#include "nmd.h"
#include "nmd_impl.h"

/*! \page componentfunctions Metadata component manipulation

  Categories have components, much like metadata objects have
  categories. Most of the component functions work on a 
  metadata object, and specify both category and component name.

  See
  NMDCategoryAllocateNewComponent(), NMDCategoryGetOrCreateComponent(),
  NMDObjectHasCategoryComponent(), NMDCategoryTryGetComponent(),
  NMDCategoryGetComponent(), 
  NMDGetCategoryIGetComponents()
 */

#undef __FUNCT__
#define __FUNCT__ "NMDAllocateComponent"
/*! An internal routine that only allocates the component 
  data structure
*/
static NMDErrorCode NMDAllocateComponent(NMD_metadata_item *rcmp)
{
  NMD_metadata_item cmp;
  NMD_MALLOC(cmp,1,struct NMD_metadata_item_,"component");
  cmp->cookie = NMDCOOKIE; cmp->set = NMDFalse;
  *rcmp = cmp;  
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCategoryAllocateNewComponent"
/*! Create a new component by name in an existing category object.
  If a component pointer is supplied, the new component object is returned,
  but this pointer is allowed to be NULL.
 */
NMDErrorCode NMDCategoryAllocateNewComponent
(NMD_metadata_category cat,const char *cmp,NMDDataType type,NMD_metadata_item *rcpt)
{
  NMD_metadata_item cpt; int idx; NMDErrorCode ierr;
  /*
   * if we are about to overflow, 
   * reallocate the components in this category
   */
  if (cat->ncmp >= cat->alloc) {
    NMD_metadata_item* newcmps; int newalloc;
    newalloc = 2*cat->alloc;
    ierr = PetscMalloc
      (newalloc*sizeof(struct NMD_metadata_item_),&newcmps); CHKERRQ(ierr);
    for (idx=0; idx<cat->ncmp; idx++)
      newcmps[idx] = cat->cmps[idx];
    ierr = PetscFree(cat->cmps); CHKERRQ(ierr);
    cat->cmps = newcmps; cat->alloc = newalloc;
  }
  /*
   * with space guaranteed, create the component
   */
  idx = cat->ncmp++;
  ierr = NMDAllocateComponent(&cpt); NMD_ERR_RETURN(ierr);
  NMD_STRDUP(cmp,cpt->name); cpt->t = type; cpt->set = NMDFalse;
  cat->cmps[idx] = cpt;
  if (rcpt) *rcpt = cpt;
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDComponentDestroy"
NMDErrorCode NMDComponentDestroy(NMD_metadata_item cmp)
{
  switch (cmp->t) {
  case NMDIntarray :
    if (cmp->ii) {
      struct NMD_intarray_struct *arraystruct =
	(struct NMD_intarray_struct *)(cmp->ii);
      if (arraystruct->length>0 && arraystruct->unique==1)
	NMD_FREE(arraystruct->data);
      NMD_FREE(cmp->ii);
    } ; 
    break;
  case NMDRealarray :
    if (cmp->rr) {
      struct NMD_realarray_struct *arraystruct =
	(struct NMD_realarray_struct *)(cmp->rr);
      if (arraystruct->length>0 && arraystruct->unique==1)
	NMD_FREE(arraystruct->data);
      NMD_FREE(cmp->rr);
    } ;
    break;
  case NMDString : NMD_FREE(cmp->cc);
    break;
  default : /* other components need no destructor */
    break;
  }
  NMD_FREE(cmp->name);
  NMD_FREE(cmp);
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCategoryGetOrCreateComponent"
/*! Retrieve a component, creating it if it doesn't already exist.
*/
NMDErrorCode NMDCategoryGetOrCreateComponent
(NMD_metadata_category cat,const char *cmp,NMDDataType type,NMD_metadata_item *cpt)
{
  NMD_metadata_item cnew; NMDTruth flg; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(cat);
  ierr = NMDCategoryTryGetComponent(cat,cmp,&cnew,&flg); NMD_ERR_RETURN(ierr);
  if (flg) {
    if (type!=cnew->t)
      NMD_ERR_REPORTii
	("Type incompatibility: old/new",(int)(cnew->t),(int)type);
    if (cpt) *cpt = cnew;
  } else {
    ierr = NMDCategoryAllocateNewComponent
      (cat,cmp,type,cpt); NMD_ERR_RETURN(ierr);
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDObjectEnsureCategoryComponent"
/* Make sure an NMD object has a specified category and component.
   If the component already exists, a type check is performed.
   A truth parameter returns whether the component was allocated
   by this routine.
*/
NMDErrorCode NMDObjectEnsureCategoryComponent
(NMD_metadata obj,const char *cat,const char *cmp,NMDDataType type,NMDTruth *nnew)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt;
  NMDTruth flg; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectTryGetCategory(obj,cat,&ctg,&flg); NMD_ERR_RETURN(ierr);
  if (!flg) {
    ierr = NMDObjectAllocateNewCategory(obj,cat,&ctg); NMD_ERR_RETURN(ierr);
  }
  ierr = NMDCategoryTryGetComponent(ctg,cmp,&cpt,&flg); NMD_ERR_RETURN(ierr);
  if (flg) {
    if (nnew) *nnew = NMDFalse;
    if (type!=cpt->t)
      NMD_ERR_REPORTii
	("Type incompatibility: old/new",(int)(cpt->t),(int)type);
  } else {
    if (nnew) *nnew = NMDTrue;
    ierr = NMDCategoryAllocateNewComponent
      (ctg,cmp,type,NULL); NMD_ERR_RETURN(ierr);
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDObjectHasCategoryComponent"
NMDErrorCode NMDObjectHasCategoryComponent
(NMD_metadata obj,const char *cat,const char *cmp,NMDTruth *f)
{
  NMD_metadata_category ctg; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectTryGetCategory(obj,cat,&ctg,f); NMD_ERR_RETURN(ierr);
  if (*f) {
    ierr = NMDCategoryTryGetComponent(ctg,cmp,NULL,f); NMD_ERR_RETURN(ierr);
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCategoryTryGetComponent"
/*! Test whether a metadata category has a certain component.
*/
NMDErrorCode NMDCategoryTryGetComponent
(NMD_metadata_category cat,const char *cmp, NMD_metadata_item *rcpt,NMDTruth *f)
{
  int icmp;
  CHECKHASNMDCOOKIE(cat);
  *f = NMDFalse;
  for (icmp=0; icmp<cat->ncmp; icmp++) {
    NMD_metadata_item cpt = cat->cmps[icmp];
    CHECKHASNMDCOOKIE(cpt);
    if (strcmp(cmp,cpt->name)==0) {
      *f = NMDTrue;
      if (rcpt) *rcpt = cpt;
      break;
    }
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCategoryGetComponents"
/*! Get a list of all component names and types in a category.
  All three output arguments are optional. The names and types arrays
  are allocated and should be freed by the user by NMD_FREE(). The names
  in the name array points to the strings in the database object, so they
  do not need to be freed.
*/
NMDErrorCode NMDCategoryGetComponents
(NMD_metadata obj,const char* cat,int *ncmp,const char ***cmps,NMDDataType **typs)
{
  NMD_metadata_category catobj; int icmp; 
  const char **names; NMDDataType *types; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetCategory(obj,cat,&catobj); CHKERRQ(ierr);
  if (ncmp) *ncmp = catobj->ncmp;
  if (cmps) {
    NMD_MALLOC(names,catobj->ncmp,char*,"name array");
    for (icmp=0; icmp<catobj->ncmp; icmp++)
      names[icmp] = catobj->cmps[icmp]->name;
    *cmps = names;
  }
  if (typs) {
    NMD_MALLOC(types,catobj->ncmp,NMDDataType,"type array");
    for (icmp=0; icmp<catobj->ncmp; icmp++)
      types[icmp] = catobj->cmps[icmp]->t;
    *typs = types;
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCategoryGetComponent"
/*! Test whether a metadata category has a certain component.
  The component has to exist.
*/
NMDErrorCode NMDCategoryGetComponent
(NMD_metadata_category cat,const char *cmp,NMD_metadata_item *cpt)
{
  NMDTruth flg; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(cat);
  ierr = NMDCategoryTryGetComponent
    (cat,cmp,cpt,&flg); NMD_ERR_RETURN(ierr);
  if (!flg) NMD_ERR_REPORTss("Could not find component in category",cmp,cat->name);
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDGetCategoryIGetComponents"
/*! For a given category, get the number of components and their names.

  All output arguments can be NULL.
  The names array is allocated; the user needs to free it. The names themselves
  are pointers to the strings in the metadata object, so they do not need to
  be freed. The types array is also allocated and needs to be freed.
*/
NMDErrorCode NMDGetCategoryIGetComponents
(NMD_metadata obj,int icat,int *ncmp,char ***cmps,NMDDataType **typs)
{
  NMD_metadata_category cat;
  CHECKHASNMDCOOKIE(obj);
  cat = obj->cats[icat];
  if (ncmp) *ncmp = cat->ncmp;
  if (cat->ncmp && (cmps || typs) ) {
    NMDDataType *types; char **names; int icmp;
    if (cmps)
      NMD_MALLOC(names,cat->ncmp,char*,"name array");
    if (typs)
      NMD_MALLOC(types,cat->ncmp,NMDDataType,"name array");
    for (icmp=0; icmp<cat->ncmp; icmp++) {
      if (cmps) names[icmp] = cat->cmps[icmp]->name;
      if (typs) types[icmp] = cat->cmps[icmp]->t;
    }
    if (cmps) *cmps = names; if (typs) *typs = types;
  }
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDComponentSetValue"
NMDErrorCode NMDComponentSetValue(NMD_metadata_item cpt,NMDDataType t,void *v)
{
  CHECKHASNMDCOOKIE(cpt);
  if (t!=cpt->t) NMD_ERR_REPORT("Type conflict");
  switch (cpt->t) {
  case NMDInt : cpt->i = *(int*)v; break;
  case NMDReal : cpt->r = *(double*)v; break;
  case NMDString : NMD_STRDUP((char*)v,cpt->cc); break;
  case NMDIntarray : 
  case NMDRealarray :
    NMD_ERR_REPORT("Please use NMDSetArrayValue"); break;
  default : NMD_ERR_REPORTi("Unknown type",(int)t); break;
  }
  cpt->set = NMDTrue;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDComponentUnsetValue"
NMDErrorCode NMDComponentUnsetValue(NMD_metadata_item cpt)
{
  CHECKHASNMDCOOKIE(cpt);
  switch (cpt->t) {
  case NMDInt :
  case NMDReal : break;
  case NMDString : NMD_FREE(cpt->cc); break;
  case NMDIntarray : {
    struct NMD_intarray_struct* iistruct = cpt->ii;
    NMD_FREE(iistruct->data);
    NMD_FREE(iistruct);
    break;}
  case NMDRealarray : {
    struct NMD_realarray_struct* rrstruct = cpt->rr;
    NMD_FREE(rrstruct->data);
    NMD_FREE(rrstruct);
    break;}
  default : NMD_ERR_REPORTi("Unknown type",cpt->t); break;
  }
  cpt->set = NMDFalse;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDComponentSetArrayValue"
NMDErrorCode NMDComponentSetArrayValue
(NMD_metadata_item cpt,NMDDataType t,void *v,int l)
{
  NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(cpt);
  switch (cpt->t) {
  case NMDInt : 
  case NMDReal : 
  case NMDString : 
    ierr = NMDComponentSetValue(cpt,cpt->t,v); NMD_ERR_RETURN(ierr);
    break;
  case NMDIntarray : {
    struct NMD_intarray_struct* iistruct;
    NMD_MALLOC(iistruct,1,struct NMD_intarray_struct,"int array struct");
    iistruct->unique = 0; iistruct->length = l;
    iistruct->data = *(int**)v;
    cpt->ii = iistruct;
    break;}
  case NMDRealarray : {
    struct NMD_realarray_struct* rrstruct;
    NMD_MALLOC(rrstruct,1,struct NMD_realarray_struct,"real array struct");
    rrstruct->unique = 0; rrstruct->length = l;
    rrstruct->data = *(NMDRealtype**)v;
    cpt->rr = rrstruct;
    break;}
  default :
    NMD_ERR_REPORTi("Unknown type",(int)t)
  }
  cpt->set = NMDTrue;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCopyItemValues"
/*! Copy data between two item structures. If the original has unique
  data, so does the clone.
*/
PetscErrorCode NMDCopyItemValues(NMD_metadata_item src,NMD_metadata_item tar)
{
  tar->set = src->set; if (!src->set) goto done;
  switch (src->t) {
  case NMDInt : tar->i = src->i; break;
  case NMDReal : tar->r = src->r ; break;
  case NMDString : NMD_STRDUP(src->cc,tar->cc); break;
  case NMDIntarray :
    {
      struct NMD_intarray_struct *tarstruct,*srcstruct = src->ii;
      int length=srcstruct->length;
      NMD_MALLOC(tarstruct,1,struct NMD_intarray_struct,"copied int array");
      tarstruct->length = length; tarstruct->unique = srcstruct->unique;
      if (srcstruct->unique) {
	int *a1=srcstruct->data,*a2; int i;
	NMD_MALLOC(a2,length,int,"int array data");
	for (i=0; i<length; i++) a2[i] = a1[i];
	tarstruct->data = a2;
      } else
	tarstruct->data = srcstruct->data;
      tar->ii = tarstruct;
    } ;
    break;
  case NMDRealarray :
    {
      struct NMD_realarray_struct *tarstruct,*srcstruct = src->rr;
      int length=srcstruct->length;
      NMD_MALLOC(tarstruct,1,struct NMD_realarray_struct,"copied real array");
      tarstruct->length = length; tarstruct->unique = srcstruct->unique;
      if (srcstruct->unique) {
	NMDRealtype *a1=srcstruct->data,*a2; int i;
	NMD_MALLOC(a2,length,NMDRealtype,"real array data");
	for (i=0; i<length; i++) a2[i] = a1[i];
	tarstruct->data = a2;
      } else
	tarstruct->data = srcstruct->data;
      tar->rr = tarstruct;
    } ;
    break;
  default : NMD_ERR_REPORTi("Can not copy items of type",(int)(src->t));
    break;
  }
 done:
  CHKMEMQ;
  return 0;
}

