/*! \file approximating.c \ingroup linear */

/*! \page approximation Approximate the coefficient matrix

  A preconditioner need not be derived from the coefficient matrix.
  For instance, in the case of a higher order finite element matrix,
  incomplete factorization preconditioner are better derived from a 
  linear element discretization of the same problem, since this matrix
  will be an M-matrix.

  This preprocessor can perform the following approximations:
  - symmetric: take the symmetric part of the coefficient matrix
  - gustafsson: apply the Gustafsson modified element matrix transformation
  (see reference [GUS] below).

\verbatim
[GUS]
@article{Gu:modified_element,
author = {Ivar Gustafsson},
title = {An Incomplete Factorization Preconditioning Method
                based on Modification of Element Matrices},
journal = {BIT},
year = {1996},
volume = {36},
pages = {86--100}
}
\endverbatim
 */
#include <stdlib.h>
#include <stdio.h>
#include "petsc.h"
#include "petscis.h"
#include "petsc.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"
#include "linear_impl.h"

#define PREPROCESSOR "approximation"

#undef __FUNCT__
#define __FUNCT__ "MatSymmetricPart"
static PetscErrorCode MatSymmetricPart
(NumericalProblem inproblem,NumericalProblem outproblem)
{
  LinearSystem problem = (LinearSystem)inproblem,
    newproblem = (LinearSystem)outproblem;
  Mat A,At,B; int first,last,nun,i; PetscScalar half=0.5,zero=0.,one=1.;
  PetscScalar avg; PetscBool flg; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = LinearSystemGetParts
    (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  ierr = MatTranspose(A,MAT_INITIAL_MATRIX,&At); CHKERRQ(ierr);
  ierr = MatDuplicate(A,MAT_COPY_VALUES,&B); CHKERRQ(ierr);
  ierr = SysProComputeQuantity
    (inproblem,"structure","n-struct-unsymm",(void*)&nun,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg && nun==0) {
    ierr = MatAXPY(B,one,At,SAME_NONZERO_PATTERN); CHKERRQ(ierr);
  } else {
    ierr = MatAXPY(B,one,At,DIFFERENT_NONZERO_PATTERN); CHKERRQ(ierr);
  }
  ierr = MatScale(B,half); CHKERRQ(ierr);
  for (i=first; i<last; i++) {
    ierr = MatSetValues(B,1,&i,1,&i,&zero,ADD_VALUES); CHKERRQ(ierr);
  }
  ierr = MatDestroy(&At); CHKERRQ(ierr);
  ierr = SysProRetrieveQuantity
    (inproblem,"variance","diagonal-average",(void*)&avg,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg) {
    int first,last,row;
    avg *= 1.e-12;
    ierr = MatGetOwnershipRange(B,&first,&last); CHKERRQ(ierr);
    for (row=first; row<last; row++) {
      ierr = MatSetValues
	(B,1,&row,1,&row,&avg,ADD_VALUES); CHKERRQ(ierr);
    }
  }
  ierr = MatAssemblyBegin(B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = LinearSystemSetParts
    (newproblem,PETSC_NULL,B,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "MatGustafssonMod"
static PetscErrorCode MatGustafssonMod
(NumericalProblem inproblem,NumericalProblem outproblem)
{
  LinearSystem problem = (LinearSystem)inproblem,
    newproblem = (LinearSystem)outproblem;
  Mat A,B; PetscScalar *a; int first,last,row,ierr;
  PetscFunctionBegin;

  ierr = LinearSystemGetParts
    (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  ierr = MatDuplicate(A,MAT_COPY_VALUES,&B); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(B,&first,&last); CHKERRQ(ierr);
  for (row=first; row<last; row++) {
    int icol,ncols; const int *cols; const PetscScalar *vals;
    ierr = MatGetRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
    for (icol=0; icol<ncols; icol++) {
      if (vals[icol].real()<0) {
	PetscScalar v = -vals[icol];
	ierr = MatSetValues
	  (B,1,&row,1,cols+icol,&v,ADD_VALUES); CHKERRQ(ierr);
	ierr = MatSetValues
	  (B,1,&row,1,&row,&v,ADD_VALUES); CHKERRQ(ierr);
      }
    }
    ierr = MatRestoreRow(A,row,&ncols,&cols,&vals); CHKERRQ(ierr);
  }
  ierr = MatAssemblyBegin(B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  ierr = MatAssemblyEnd(B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);

  {
    /*
     * Enforce non-negative diagonal
     */
    Vec V;
    ierr = VecCreateSeq(MPI_COMM_SELF,last-first,&V); CHKERRQ(ierr);
    ierr = MatGetDiagonal(B,V); CHKERRQ(ierr);
    ierr = VecGetArray(V,&a); CHKERRQ(ierr);
    for (row=first; row<last; row++) {
      if (a[row-first].real()<0) {
	PetscScalar v=0;
	ierr = MatSetValues(B,1,&row,1,&row,&v,INSERT_VALUES); CHKERRQ(ierr);
      }
    }
    ierr = VecRestoreArray(V,&a); CHKERRQ(ierr);
    ierr = VecDestroy(&V); CHKERRQ(ierr);
    ierr = MatAssemblyBegin(B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
    ierr = MatAssemblyEnd(B,MAT_FINAL_ASSEMBLY); CHKERRQ(ierr);
  }
  ierr = LinearSystemSetParts
    (newproblem,PETSC_NULL,B,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_approximation_choices"
static PetscErrorCode setup_approximation_choices()
{
  SalsaTransform tf; SalsaTransformObject cur; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformGetByName(PREPROCESSOR,&tf); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"none",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"use coefficient matrix"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"symmetric",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"symmetric part"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"gustafsson",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"gustafsson M-matrix"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "specific_approximation_choices"
/*!
  This is the 'specific setup' phase of the approximation preprocessor.
  See \ref modes for details.

  This routine  eliminates the Gustafsson approximation for diagonally dominant
  systems, and the symmetric for symmetric systems.
*/
static PetscErrorCode specific_approximation_choices
(NumericalProblem inproblem,SalsaTransform transform)
{
  LinearSystem problem = (LinearSystem)inproblem; Mat A;
  PetscReal dd,anorm; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSystemGetParts
    (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);

  /*
   * No Gustafsson approximation for diagonally dominant matrices
   */
  ierr = SysProRetrieveQuantity
    (inproblem,"simple","diagonal-dominance",(void*)&dd,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg && dd>=0.) {
    SalsaTransformObject tf;
    ierr = TransformObjectGetByName
      (PREPROCESSOR,"gustafsson",&tf); CHKERRQ(ierr);
    ierr = TransformObjectMark(tf); CHKERRQ(ierr);
  }

  /*
   * symmetric approximations of a symmetric matrices
   * are a silly idea
   */
  ierr = SysProRetrieveQuantity
    (inproblem,"simple","symmetry-anorm",(void*)&anorm,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg && anorm==0.) {
    SalsaTransformObject tf;
    ierr = TransformObjectGetByName
      (PREPROCESSOR,"symmetric",&tf); CHKERRQ(ierr);
    ierr = TransformObjectMark(tf); CHKERRQ(ierr);
  }


  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "approximate_system"
static PetscErrorCode approximate_system
(const char *type,int nopt,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx, PetscBool *success)
{
  LinearSystem problem = (LinearSystem)inproblem, newproblem;
  PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
 
  {
    Mat A,B;
    ierr = LinearSystemGetParts
      (problem,&A,&B,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
    if (A!=B) SETERRQ(PETSC_COMM_SELF,1,"A and B should be the same before approximation");
  }
  ierr = LinearSystemDuplicatePointers(problem,&newproblem); CHKERRQ(ierr);

  *success = PETSC_TRUE;

  ierr = PetscStrcmp("none",type,&flg); CHKERRQ(ierr);
  if (flg) {
    goto done;
  }
  
  ierr = PetscStrcmp("symmetric",type,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatSymmetricPart(inproblem,(NumericalProblem)newproblem); CHKERRQ(ierr);
    goto done;
  }
  
  ierr = PetscStrcmp("gustafsson",type,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatGustafssonMod(inproblem,(NumericalProblem)newproblem); CHKERRQ(ierr);
    goto done;
  }

  SETERRQ1(PETSC_COMM_SELF,1,"Unknown matrix approximation <%s>",type);

 done:
  *outproblem = (NumericalProblem)newproblem;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "unapproximate_system"
static PetscErrorCode unapproximate_system
(const char *type,PetscBool overwrite,void *gctx,void *ctx,
 NumericalProblem problem,NumericalProblem nextproblem,
 NumericalSolution before,NumericalSolution after)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSolutionCopy
    ((LinearSolution)before,(LinearSolution)after); CHKERRQ(ierr);
  /*ierr = LinearSolutionDelete((LinearSolution)before); CHKERRQ(ierr);*/
  //ierr = DeleteLinearSystem((LinearSystem)problem); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareApproximationPreprocessor"
PetscErrorCode DeclareApproximationPreprocessor(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DeclarePreprocessor
    ("approximation",
     &setup_approximation_choices,&specific_approximation_choices,0,0,
     0,0,
     &approximate_system,&unapproximate_system); CHKERRQ(ierr);
  ierr = PreprocessorSetPreservedCategories
    (PREPROCESSOR,"laplace"); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
