/*! \page ksp The iterative method

  The iterative method is not really a transformation, but it is
  the last choice made in a preprocessor loop before the final
  solver is called. This means that no new matrix analysis is performed 
  after applying this transformation.
 */

#include <stdlib.h>
#include <stdio.h>
#include "string.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"
#include "sysprosuit.h"
#include "anamod.h"
#include "linksp.h"
#include "petscmat.h"
#include "petscpc.h"
#include "petscksp.h"

#define PREPROCESSOR "ksp"
int gmrescycleid;

#undef __FUNCT__
#define __FUNCT__ "is_gmres_method"
static PetscErrorCode is_gmres_method(KSPType kspt,PetscBool *f)
{
  SalsaTransformObject ksp; int is; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformObjectGetByName("ksp",(char*)kspt,&ksp); CHKERRQ(ierr);
  ierr = TransformObjectGetIntAnnotation(ksp,"is-gmres",&is,f); CHKERRQ(ierr);
  *f = TRUTH(*f && is==1);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_ksp_choices"
static PetscErrorCode setup_ksp_choices()
{
  SalsaTransform ksp; SalsaTransformObject cur;
  PetscErrorCode ierr;

  PetscFunctionBegin;

  ierr = PetscObjectComposedDataRegister(&gmrescycleid); CHKERRQ(ierr);
  ierr = TransformGetByName(PREPROCESSOR,&ksp); CHKERRQ(ierr);
  ierr = SetSuitabilityContextTester(ksp,&AnaModCheckValidFeatureSet); CHKERRQ(ierr);

  /* symmetric substitutions */
  ierr = SysProDefineCharAnnotation(PREPROCESSOR,"symm"); CHKERRQ(ierr);
  /* requires transpose */
  ierr = SysProDefineIntAnnotation(PREPROCESSOR,"trans"); CHKERRQ(ierr);
  /* has restart stuff */
  ierr = SysProDefineIntAnnotation(PREPROCESSOR,"is-gmres"); CHKERRQ(ierr);

  /*
   * Define our repertoire of KSPs
   */

  /*
   * GMRES
   */
  ierr = NewTransformObject(PREPROCESSOR,KSPGMRES,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"Generalized Minimum Residual Method"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"is-gmres",1); CHKERRQ(ierr);
  ierr = TransformObjectDefineOption(cur,"restart length"); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,5); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,20); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,50); CHKERRQ(ierr);

  /*
   * fGMRES
   */
  ierr = NewTransformObject(PREPROCESSOR,KSPFGMRES,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"flexible GMRES"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"is-gmres",1); CHKERRQ(ierr);
  ierr = TransformObjectDefineOption(cur,"restart length"); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,5); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,20); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,50); CHKERRQ(ierr);
  
  /*
   * lGMRES
   */
  ierr = NewTransformObject(PREPROCESSOR,KSPLGMRES,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"loose GMRES"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"is-gmres",1); CHKERRQ(ierr);
  ierr = TransformObjectDefineOption(cur,"restart length"); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,5); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,20); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,50); CHKERRQ(ierr);
  
  /* other methods in no particular order */
  ierr = NewTransformObject(PREPROCESSOR,KSPTFQMR,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"Transpose-free QMR"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,KSPCGS,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"Conjugate Gradients Squared"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,KSPBICG,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"Bi-Conjugate Gradients"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"trans",1); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,KSPBCGS,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"Bi-conjugate Gradients Stabilized"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,KSPBCGSL,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"Bi-conjugate Gradients Stabilized(l)"); CHKERRQ(ierr);
  ierr = TransformObjectDefineOption(cur,"gmres length"); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,2); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,3); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,4); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,5); CHKERRQ(ierr);
  ierr = TransformObjectAddOption(cur,6); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,KSPCGNE,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"CG on the normal equations"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"trans",1); CHKERRQ(ierr);
  
  ierr = NewTransformObject(PREPROCESSOR,KSPCG,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"plain conjugate gradients"); CHKERRQ(ierr);
  {
    FeatureSet symmetry;
    ierr = NewFeatureSet(&symmetry); CHKERRQ(ierr);
    ierr = AddToFeatureSet
      (symmetry,"simple","symmetry-snorm",PETSC_NULL); CHKERRQ(ierr);
    ierr = AddToFeatureSet
      (symmetry,"simple","symmetry-anorm",PETSC_NULL); CHKERRQ(ierr);
    ierr = TransformObjectSetSuitabilityFunction
      (cur,(void*)symmetry,&onlyforsymmetricproblem); CHKERRQ(ierr);
  }

  ierr = NewTransformObject(PREPROCESSOR,KSPPREONLY,&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"Preconditioner application"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "unset_ksps"
static PetscErrorCode unset_ksps()
{
  SalsaTransformObject to; void *ctx;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TransformObjectGetByName(PREPROCESSOR,KSPCG,&to); CHKERRQ(ierr);
  ierr = TransformObjectGetSuitabilityFunction
    (to,&ctx,PETSC_NULL); CHKERRQ(ierr);
  ierr = DeleteFeatureSet((FeatureSet)ctx); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "disable_ksps"
static PetscErrorCode disable_ksps
(NumericalProblem theproblem,SalsaTransform ksp)
{
  SalsaTransformObject curpc;
  PetscBool has_annot,flg; PetscErrorCode ierr;
  PetscFunctionBegin;

  {
    const char *pct; int pcv;
    ierr = PreprocessorGetSetting("pc",&pct,&pcv); CHKERRQ(ierr);
    ierr = TransformObjectGetByName("pc",pct,&curpc); CHKERRQ(ierr);
  }

  /* we use all KSPs .... */
  ierr = TransformObjectsUnmarkAll(ksp); CHKERRQ(ierr);

  /* ... except if the current PC has no transpose ... */
  ierr = TransformObjectGetIntAnnotation
    (curpc,"notrans",(int*)&flg,&has_annot); CHKERRQ(ierr);
  if (flg && has_annot) {
    int iksp,n; SalsaTransformObject *objs;
    ierr = TransformGetObjects(ksp,&n,&objs); CHKERRQ(ierr);
    for (iksp=0; iksp<n; iksp++) {
      /* ... then we mark the KSPs that need a transpose */
      SalsaTransformObject to = objs[iksp]; int trans;
      ierr = TransformObjectGetIntAnnotation
	(to,"trans",&trans,&flg); CHKERRQ(ierr);
      if (flg && trans) {
	ierr = TransformObjectMark(to); CHKERRQ(ierr);
      }
    }
  }

  /* ... and if the pc is LU, then we use only preonly and vv */
  ierr = TransformObjectGetIntAnnotation
    (curpc,"directsolve",(int*)&flg,&has_annot); CHKERRQ(ierr);
  if (has_annot && flg) {
    int iksp,n; SalsaTransformObject *objs;
    ierr = TransformGetObjects(ksp,&n,&objs); CHKERRQ(ierr);
    for (iksp=0; iksp<n; iksp++) {
      /* ... and we mark everything but preonly  */
      SalsaTransformObject to = objs[iksp]; const char *name;
      ierr = TransformObjectGetName(to,&name); CHKERRQ(ierr);
      if (strcmp(name,KSPPREONLY)) {
	ierr = TransformObjectMark(to); CHKERRQ(ierr);
      }
    }
  } else {
    /* if it's not a direct solve */
    int iksp,n; SalsaTransformObject *objs;
    ierr = TransformGetObjects(ksp,&n,&objs); CHKERRQ(ierr);
    for (iksp=0; iksp<n; iksp++) {
      /* we disable preonly */
      SalsaTransformObject to = objs[iksp]; const char *name;
      ierr = TransformObjectGetName(to,&name); CHKERRQ(ierr);
      if (!strcmp(name,KSPPREONLY)) {
	ierr = TransformObjectMark(to); CHKERRQ(ierr);
      }
    }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "set_ksp_options"
static PetscErrorCode set_ksp_options(SalsaTransformObject tf, int kspv)
{
  KSPType kspt;;
  int g; PetscBool f; PetscErrorCode ierr; const char *opt;

  PetscFunctionBegin;

  ierr = TransformObjectGetName(tf,(const char**)&kspt); CHKERRQ(ierr);

  opt = "-ksp_type";
  ierr = PetscOptionsHasName(PETSC_NULL,opt,&f); CHKERRQ(ierr);
  if (f) {ierr = PetscOptionsClearValue(opt); CHKERRQ(ierr);}
  ierr = PetscOptionsSetValue(opt,kspt); CHKERRQ(ierr);

  ierr = TransformObjectGetIntAnnotation(tf,"is-gmres",&g,&f); CHKERRQ(ierr);
  if (f && g) {
    char val[5];
    sprintf(val,"%d",kspv);
    opt = "-ksp_gmres_restart";
    ierr = PetscOptionsHasName(PETSC_NULL,opt,&f); CHKERRQ(ierr);
    if (f) {ierr = PetscOptionsClearValue(opt); CHKERRQ(ierr);}
    ierr = PetscOptionsSetValue(opt,val); CHKERRQ(ierr);
    goto method_set;
  }

  ierr = PetscStrcmp(kspt,KSPBCGSL,&f); CHKERRQ(ierr);
  if (f) {
    char val[5];
    sprintf(val,"%d",kspv);
    opt = "-ksp_bcgsl_ell";
    ierr = PetscOptionsHasName(PETSC_NULL,opt,&f); CHKERRQ(ierr);
    if (f) {ierr = PetscOptionsClearValue(opt); CHKERRQ(ierr);}
    ierr = PetscOptionsSetValue(opt,val); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ksp_bcgsl_xres","0"); CHKERRQ(ierr);
    goto method_set;
  }

 method_set:
  /*
   * GMRES only works with preconditioned norm
   */
  ierr = PetscStrcmp(kspt,KSPGMRES,&f); CHKERRQ(ierr);
  if (f) {
    ierr = PetscOptionsSetValue
      ("-ksp_norm_type","preconditioned"); CHKERRQ(ierr);
  } else {
    ierr = PetscOptionsSetValue
      ("-ksp_norm_type","unpreconditioned"); CHKERRQ(ierr);
  }
  /*
   * FGMRES only works with right pc
   */
  ierr = PetscOptionsClearValue("-ksp_left_pc"); CHKERRQ(ierr);
  ierr = PetscOptionsClearValue("-ksp_right_pc"); CHKERRQ(ierr);
  ierr = PetscOptionsClearValue("-ksp_symmetric_pc"); CHKERRQ(ierr);
  ierr = PetscStrcmp(kspt,KSPFGMRES,&f); CHKERRQ(ierr);
  if (f) {
    ierr = PetscOptionsSetValue("-ksp_right_pc",PETSC_NULL); CHKERRQ(ierr);
  } else {
    ierr = PetscOptionsSetValue("-ksp_left_pc",PETSC_NULL); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_ksp"
static PetscErrorCode setup_ksp
(const char* kspt,int kspv,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx,PetscBool *success)
{
  LinearSystem newproblem,problem = (LinearSystem)inproblem;
  SalsaTransformObject ksp; KSP method; PetscBool isgmres; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TransformObjectGetByName("ksp",(KSPType)kspt,&ksp); CHKERRQ(ierr);
  ierr = set_ksp_options(ksp,kspv); CHKERRQ(ierr);

  /* get the ksp context which was created by the pc preprocessor */
  ierr = PreprocessorGetContext("pc",(void**)&method); CHKERRQ(ierr);
#if 0
  {
    PC pc; MPI_Comm comm;
    ierr = PreprocessorGetContext("pc",(void**)&pc); CHKERRQ(ierr);
    ierr = PetscObjectGetComm((PetscObject)pc,&comm); CHKERRQ(ierr);
    ierr = KSPCreate(comm,&method); CHKERRQ(ierr);
    ierr = SysProLinearInstallCustomKSPMonitor(method); CHKERRQ(ierr);
    /*ierr = KSPSetPC(method,pc); CHKERRQ(ierr);*/
    *(KSP*)ctx = method;
  }
#endif
  ierr = KSPSetFromOptions(method); CHKERRQ(ierr);
  ierr = is_gmres_method((KSPType)kspt,&isgmres); CHKERRQ(ierr);
  if (isgmres) {
    ierr = PetscObjectComposedDataSetInt
      ((PetscObject)method,gmrescycleid,kspv); CHKERRQ(ierr);
  } else {
    ierr = PetscObjectComposedDataSetInt
      ((PetscObject)method,gmrescycleid,-1); CHKERRQ(ierr);
  }
  ierr = PetscPushErrorHandler(&PetscReturnErrorHandler,NULL); CHKERRQ(ierr);
  ierr = KSPSetUp(method);
  if (ierr) {
    *success = PETSC_FALSE ; 
  } else {
    *success = PETSC_TRUE;
    ierr = LinearSystemDuplicatePointers(problem,&newproblem); CHKERRQ(ierr);
    *outproblem = (NumericalProblem)newproblem;
  }
  ierr = PetscPopErrorHandler(); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "unset_ksp"
static PetscErrorCode unset_ksp
(const char *kspt,PetscBool overwrite,void *gctx,void *ctx,
 NumericalProblem inproblem,NumericalProblem nextproblem,
 NumericalSolution old,NumericalSolution nnew)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
#if 0
  {
    KSP solver = (KSP)ctx;
    ierr = KSPDestroy(solver); CHKERRQ(ierr);
  }
#endif
  ierr = LinearSolutionCopy
    ((LinearSolution)old,(LinearSolution)nnew); CHKERRQ(ierr);
  /*ierr = LinearSolutionDelete((LinearSolution)old); CHKERRQ(ierr);*/
  //ierr = DeleteLinearSystem((LinearSystem)inproblem); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareKSPPreprocessor"
PetscErrorCode DeclareKSPPreprocessor(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DeclarePreprocessor
    ("ksp",
     &setup_ksp_choices,/* this_preprocessor_setup */
     &disable_ksps, /* specific_setup */
     0, /* specific unset */
     &unset_ksps, /* global_unset */
     0,0, /* context create/delete */
     &setup_ksp,&unset_ksp /* start/end function */
     ); CHKERRQ(ierr);
  ierr = PreprocessorSetPreservedCategories
    (PREPROCESSOR,"laplace"); CHKERRQ(ierr);
  /*
  ierr = DeclarePreprocessorIntelligentChoice
    ("scaling",&set_intelligent_scaling); CHKERRQ(ierr);
  */
  PetscFunctionReturn(0);
}

PetscErrorCode (*custommonitor)(KSP,int,PetscReal,void*) = NULL;
void *monitordata = NULL;

#undef __FUNCT__
#define __FUNCT__ "SysProLinearInstallCustomKSPMonitor"
PetscErrorCode SysProLinearInstallCustomKSPMonitor(KSP solver)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (custommonitor) {
    ierr = KSPMonitorSet
      (solver,custommonitor,monitordata,PETSC_NULL); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProLinearDeclareCustomKSPMonitor"
PetscErrorCode SysProLinearDeclareCustomKSPMonitor
(PetscErrorCode(*monitor)(KSP,int,PetscReal,void*),void *data)
{
  PetscFunctionBegin;
  custommonitor = monitor;
  if (data) monitordata = data;
  PetscFunctionReturn(0);
}

