/*! \mainpage SysPro: a System preProcessors library


  \section Introduction

  This is a library of preprocessors for numerical problems, that is,
  mappings of one numerical problem into another, presumably more simple one,
  of the same type. For example, scaling a linear system.
  The SysPro library operates in two modes:
  - exhaustive mode: all possibly choices of a preprocessor are explored
  in sequence; this mode is set by command line options.
  - intelligent mode: based on problem properties, a suitable choice for
  each preprocessor is made.

  See \ref modes for more details.

  Each preprocessor has the following structure, which is executed
  in the PreprocessedSolution() routine:
  - a global setup is performed. This is a good place for computing
  problem features with the AnaMod library.
  - a specific setup is performed; this can for instance disable certain
  preprocessor choices based on the computed problem features.
  - a selection is made; this can be
    - a first choice, if the preprocessor is applied in exhaustive mode
    - an intelligent choice, if the exhaustive mode is off, and an 
    intelligent choice routine exists
    - some default choice otherwise

  In case of exhaustive mode, the following steps are executed inside a loop
  over all choices for this preprocessor, and possible all numerical option
  settings:
  - the start function transforms the problem into a preprocessed problem
  - if a next preprocessor is defined, it is applied;
  otherwise, the problem solving routine is applied 
  (see \ref declaration).
  - the end function backtransforms the solution of the preprocessed problem
  into that of the original problem.

  \section Topics

  \subpage use

  \subpage reporting and \subpage tracing

  \subpage linear

  \subpage external

  \subpage options

  \author Victor Eijkhout
  \version 1.3
  \date unreleased

  \section diary Change log

  1.3
  2008/08/20 : DeclarePreprocessor() has an extra argument for global
  deallocation at the end of a program run. Currently used in the ksp
  preprocessor.

  2008/05/10 : end function now has two NumericalProblem parameters;
  this is necessary for freeing the recursive problem.

  2007 : Accomodated the array length parameter in anamod/nmd

*/

/*! \page use Use of the SysPro package

  \section sysprosetup Setup of the SysPro system

  The setup of SysPro has to be done once per program run.

  \subsection sysprogsetup Global declarations

  All use of SysPro has to be inside calls to SysProInitialize() and
  SysProFinalize(). These allocate and deallocate global data structures.
  You could place them right next to MPI_Initialize/Finalize or 
  PetscInitialize/Finalize calls. The SysProFinalize() call can also be used
  to deallocate data that was constructed during preprocessor setup.

  Another step in the global setup of SysPro is a call to 
  SysProDeclareFunctions(), This declares functions that are of use
  to all preprocessors that will be declared later.

  \subsection declaration Preprocessor declaration

  Preprocessors are declared with calls to DeclarePreprocessor(), which 
  installs the setup functions and the forward / backward transformations.

  Further specifications can be given for a specific preprocessor:
  - DeclarePreprocessorIntelligentChoice() for installing a routine that will
  intelligently pick a preprocessor choice.
  - see \ref anamodinterface for metadata category handling.

  \section setup Setup and invocation

  After the setup as described above,
  PreprocessorsOptionsHandling() can be called to provide the user with 
  runtime control (see section \ref modes) over the workings of SysPro.

  Preprocessed problem solving is activated by a call to
  PreprocessedProblemSolving(). This causes all declared preprocessors
  to be applied in sequence. Finally, the ultimately remaining problem is
  solved with the routine declared by PreprocessorsDeclareProblemSolver().

  \section modes Usage modes

  SysPro can be used in several modes:
  - one can specify the exact preprocessor values (or several values);
  - one can specify to test exhaustively the values of one preprocessor
  or all of them;
  - SysPro can intelligently pick the appropriate preprocessor.

  The intelligent preprocessor choice uses a model where each preprocessor
  has a measure of how applicable it is; the SysPro system then picks the 
  most appropriate preprocessor from a given class. See \ref suit.

  See \ref options for details on specific and exhaustive testing.

  \section modules Predefined preprocessors

  The SysPro package comes with a number of predefined preprocessors,
  mostly geared towards linear system processing.

  \page linear

  \page lineartransform
 */

/*! \page external The interface to other packages

  SysPro by itself is not of a lot of use: it is a framework for 
  tying together transformations and operations. There is also an interface
  for computing metadata.

  \subpage anamodinterface

 */

/*! \page linear The linear systems example package
  The mechanism of preprocessed solving through forward and backward
  transformations can be applied to all sorts of numerical problems.
  At present, we only supply code for linear system solving; 
  see \ref linearfile.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "petscmat.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "syspro_impl.h"

#define NPREPROCESS 25
SystemPreprocessor *preprocessors=NULL;
const char **currentpreprocessors,**currentchoices=NULL;
int *currentoptions=NULL;
int npreprocess=0,preprocesslevel;
static void **preprocessorcontexts,*solutioncontext;
static PetscErrorCode(**unsetpreprocessor)();

struct PreprocessorsGlobalInfo_ {
  PetscErrorCode(*problemmonitor)(NumericalProblem);
  PetscErrorCode(*classstaticsetup)(const char*);
  /** This routine is executed on the creation of a new
      preprocessor. It can be used to install standard options
      in the preprocessor transform object.
  */
  PetscErrorCode(*classdynamicsetup)(const char*,NumericalProblem);
  /** This routine is invoked at the start of each preprocessor
      class. It is not supposed to contain problem-dependent
      actions. It is useful for printing trace messages, and
      performing analysis on each incoming problem.
  */      
  PetscErrorCode(*classproblemcloner)
  (const char*,const char*,int,NumericalProblem,NumericalProblem);
  /** This routine is called everytime a new problem is created with
      a class/option pair. It can be used to copy preserved metadata elements
  */
  PetscErrorCode(*computecategory)(char*,NumericalProblem);
  /** This routine is called in sequence with the names of the
      required metadata categories.
   */
  PetscErrorCode(*metadatacomputer)(char*,char*,Mat,void*,PetscBool*);
  PetscErrorCode(*clonecontext)(const char*,const char*,void*,void**);
  PetscErrorCode(*freecontext)(void*);
  PetscErrorCode(*problemsolver)(NumericalProblem,void*,NumericalSolution*);
  PetscErrorCode(*problemdelete)(NumericalProblem);
  PetscErrorCode(*errortracer)(NumericalProblem,NumericalSolution,const char*);
  PetscErrorCode(*solutioncreator)(NumericalProblem,NumericalSolution*);
  PetscErrorCode(*solutioncopy)(NumericalSolution,NumericalSolution);
  PetscErrorCode(*solutiondelete)(NumericalSolution);
  PetscErrorCode(*solutioncontextdelete)(NumericalSolution);
} ; 
typedef struct PreprocessorsGlobalInfo_* PreprocessorsGlobalInfo;
static PreprocessorsGlobalInfo GlobalInfo = NULL;

#undef __FUNCT__
#define __FUNCT__ "CreateGlobalInfo"
static PetscErrorCode CreateGlobalInfo()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (!GlobalInfo) {
    ierr = PetscMalloc
      (sizeof(struct PreprocessorsGlobalInfo_),&GlobalInfo); CHKERRQ(ierr);
    ierr = PetscMemzero
      (GlobalInfo,sizeof(struct PreprocessorsGlobalInfo_)); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProInitialize"
/*! Allocate SysPro globals. See also SysProFinalize(). */
PetscErrorCode SysProInitialize()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = CreateGlobalInfo(); CHKERRQ(ierr);
  ierr = PetscMalloc
    (NPREPROCESS*sizeof(SystemPreprocessor),&preprocessors); CHKERRQ(ierr);
  ierr = PetscMalloc
    (NPREPROCESS*sizeof(char*),&currentpreprocessors); CHKERRQ(ierr);
  ierr = PetscMalloc
    (NPREPROCESS*sizeof(void*),&preprocessorcontexts); CHKERRQ(ierr);
  ierr = PetscMalloc
    (NPREPROCESS*sizeof(char*),&currentchoices); CHKERRQ(ierr);
  ierr = PetscMalloc
    (NPREPROCESS*sizeof(int),&currentoptions); CHKERRQ(ierr);
  ierr = PetscMalloc
    (NPREPROCESS*sizeof(PetscErrorCode(*)()),&unsetpreprocessor); CHKERRQ(ierr);
  ierr = PetscMemzero
    (unsetpreprocessor,NPREPROCESS*sizeof(PetscErrorCode(*)())); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProFinalize"
PetscErrorCode SysProFinalize()
{
  int ip; PetscErrorCode ierr;
  PetscFunctionBegin;
  for (ip=0; ip<npreprocess; ip++) {
    SystemPreprocessor thisone = preprocessors[ip];
    if (unsetpreprocessor[ip]) {
      ierr = (unsetpreprocessor[ip])(); CHKERRQ(ierr);
    }
    if (thisone->preserved) {
      ierr = PetscFree(thisone->preserved); CHKERRQ(ierr);
    }
    ierr = PetscFree(thisone->name); CHKERRQ(ierr);
    ierr = DeregisterTransform(thisone->transform); CHKERRQ(ierr);
    ierr = PetscFree(thisone); CHKERRQ(ierr);
  }
  ierr = PetscFree(preprocessors); CHKERRQ(ierr);
  ierr = PetscFree(currentpreprocessors); CHKERRQ(ierr);
  ierr = PetscFree(preprocessorcontexts); CHKERRQ(ierr);
  ierr = PetscFree(currentchoices); CHKERRQ(ierr);
  ierr = PetscFree(currentoptions); CHKERRQ(ierr);
  ierr = PetscFree(unsetpreprocessor); CHKERRQ(ierr);
  ierr = PetscFree(GlobalInfo); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclarePreprocessor"
/*!
  Declare a preprocessor class, by specifing its various members.

  The \c name argument should not contain the colon character.

  Here is an explanation of the various function arguments.

  \c this_preprocessor_setup() : this routine is called only once,
  inside this function. This is a good place for defining all the 
  preprocessors in this class

  \c specific_setup(NumericalProblem,SalsaTransform) : this is called
  at the start of a preprocessing stage; one could use this for
  computing matrix metadata.

  \c global_unset(void) : this is called in SysProFinalize().

  \c ctx_create(NumericalProblem,void**) : create an object that can
  be used for the duration of the application of this preprocessor

  \c ctxdelete(void*) : delete the context again

  \c start_function : this is the function that performs the forward
  transform of the problem. Prototype:
\code
  PetscErrorCode start_function
    (char             *classmember,
     int               optionvalue,
     PetscBool        overwrite,
     NumericalProblem  problem,
     NumericalProblem *transformedproblem,
     void             *globalcontext,
     void            **localcontext,
     PetscBool       *success)
\endcode

  \c end_function : this is the backtransform. Its main task is copying
  or backtransforming the preprocessed solution to the original solution.
\code
  PetscErrorCode end_function
    (char             *classmember,
     PetscBool        overwrite,
     void             *globalcontext,
     void             *localcontext,
     NumericalProblem  pproblem,
     NumericalProblem  oproblem,
     NumericalSolution psolution,
     NumericalSolution osolution)
\endcode
where \c pproblem and \c psolution are the preprocessed quantities, 
the end function has to unprocess them and leave the result in
\c oproblem, \c osolution. Actually, \c oproblem is only for reference.
*/
PetscErrorCode DeclarePreprocessor
(const char *name,
 PetscErrorCode(*this_preprocessor_setup)(),
 PetscErrorCode(*specific_setup)(NumericalProblem,SalsaTransform),
 PetscErrorCode(*specific_unset)(NumericalProblem),
 PetscErrorCode(*global_unset)(),
 PetscErrorCode(*ctxcreate)(NumericalProblem,void**),
 PetscErrorCode(*ctxdelete)(void*),
 PetscErrorCode(*start_function)(const char*,int,PetscBool,NumericalProblem,NumericalProblem*,void*,void**,PetscBool*),
 PetscErrorCode(*end_function)(const char*,PetscBool,void*,void*,NumericalProblem,NumericalProblem,NumericalSolution,NumericalSolution)
 )
{
  SystemPreprocessor snew;
  PetscErrorCode ierr;

  PetscFunctionBegin;
  if (!preprocessors) SETERRQ(PETSC_COMM_SELF,1,"Forgotten SysProInitialize");
  if (npreprocess>=NPREPROCESS)
    SETERRQ(PETSC_COMM_SELF,1,"Can not declare further system preprocessors");
  {
    int p;
    for (p=0; p<npreprocess; p++) 
      if (!strcmp(name,preprocessors[p]->name))
	SETERRQ1(PETSC_COMM_SELF,1,"There already is a preprocessor called <%s>",name);
  }
  unsetpreprocessor[npreprocess] = global_unset;
  ierr = PetscMalloc(sizeof(struct SystemPreprocessor_),&snew); CHKERRQ(ierr);
  ierr = PetscMemzero(snew,sizeof(struct SystemPreprocessor_)); CHKERRQ(ierr);
  preprocessors[npreprocess++] = snew;

  snew->exhaustive = PETSC_FALSE;
  {
    char *name;
    ierr = PetscStrallocpy(name,&name); CHKERRQ(ierr);
    snew->name = name;
  }
  ierr = NewTransform(name,&(snew->transform)); CHKERRQ(ierr);
  if (GlobalInfo->classstaticsetup) {
    ierr = (GlobalInfo->classstaticsetup)(name); CHKERRQ(ierr);
  }
  if (!this_preprocessor_setup)
    SETERRQ(PETSC_COMM_SELF,1,"No setup function given");
  ierr = (*this_preprocessor_setup)(); CHKERRQ(ierr);
  snew->setup = specific_setup;
  snew->unset = specific_unset;
  snew->ctxcreate = ctxcreate; snew->ctxdelete = ctxdelete;
  snew->start_function = start_function; snew->end_function = end_function;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProDeclareProblemMonitor"
PetscErrorCode SysProDeclareProblemMonitor
(PetscErrorCode(*monitor)(NumericalProblem))
{
  PetscFunctionBegin;
  if (monitor) GlobalInfo->problemmonitor = monitor;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProDeclareErrorTracer"
PetscErrorCode SysProDeclareErrorTracer
(PetscErrorCode(*tracer)(NumericalProblem,NumericalSolution,const char*))
{
  PetscFunctionBegin;
  if (tracer) GlobalInfo->errortracer = tracer;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProGetErrorTracer"
PetscErrorCode SysProGetErrorTracer
(PetscErrorCode(**tracer)(NumericalProblem,NumericalSolution,const char*))
{
  PetscFunctionBegin;
  if (tracer) *tracer = GlobalInfo->errortracer;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclarePreprocessorIntelligentChoice"
/*! Install a function to pick the optimal choice for a preprocessor */
PetscErrorCode DeclarePreprocessorIntelligentChoice
(const char *name,PetscErrorCode(*picker)(NumericalProblem,const char**,const char**))
{
  SystemPreprocessor pp;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SystemPreprocessorGetByName(name,&pp); CHKERRQ(ierr);
  pp->intelligence = picker;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "NumericalProblemGetComm"
PetscErrorCode NumericalProblemGetComm(NumericalProblem prob,MPI_Comm *comm)
{
  PetscFunctionBegin;
  *comm = prob->comm;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProDeclareFunctions"
/*! Install various functions
  - \c classstaticsetup : this function is called on each processor as
  it is being created; see DeclarePreprocessor().
  - \c classdynamicsetup : this function is called as any invocation
  of a preprocessor starts; see PreprocessedSolution();
  - \c classproblemcloner : a function to clone the context : optional
  see \ref tracing for more details.
  - \c problemsolver : the ultimate problem solver : required
  - \c problemdelete : delete a problem object
  - \c solutioncreator : creates a solution object; optional, but required
  a preprocessor has an endfunction.
  - \c solutioncopy : guess what this does; optional
  - \c solutiondelete : optional, but needed if solutioncopy is used
  - \c contextcloner : problems can carry a context; this clones the context
  if a problem is copied; otherwise the pointer is simply duplicated
  - \c contextfree : used to delete cloned contexts
  - \c solutioncontextdelete : hm.

*/
PetscErrorCode SysProDeclareFunctions
(
 PetscErrorCode(*classstaticsetup)(const char*),
 PetscErrorCode(*classdynamicsetup)(const char*,NumericalProblem),
 PetscErrorCode(*classproblemcloner)(const char*,const char*,int,NumericalProblem,NumericalProblem),
 PetscErrorCode(*problemsolver)(NumericalProblem,void*,NumericalSolution*),
 PetscErrorCode(*problemdelete)(NumericalProblem),
 PetscErrorCode(*solutioncreator)(NumericalProblem,NumericalSolution*),
 PetscErrorCode(*solutioncopy)(NumericalSolution,NumericalSolution),
 PetscErrorCode(*solutiondelete)(NumericalSolution),
 PetscErrorCode(*ctxcloner)(const char*,const char*,void*,void**),
 PetscErrorCode(*ctxfree)(void*),
 PetscErrorCode(*solutioncontextdelete)(NumericalSolution)
 )
{
  PetscFunctionBegin;

  if (classstaticsetup) GlobalInfo->classstaticsetup = classstaticsetup;
  if (classdynamicsetup) GlobalInfo->classdynamicsetup = classdynamicsetup;
  if (classproblemcloner) GlobalInfo->classproblemcloner = classproblemcloner;

  if (!problemsolver) SETERRQ(PETSC_COMM_SELF,1,"A solver routine is required");
  GlobalInfo->problemsolver = problemsolver;
  GlobalInfo->problemdelete = problemdelete;

  if (ctxcloner) GlobalInfo->clonecontext = ctxcloner;
  if (ctxfree) GlobalInfo->freecontext = ctxfree;

  if (solutioncreator) GlobalInfo->solutioncreator = solutioncreator;
  if (solutioncopy) GlobalInfo->solutioncopy = solutioncopy;
  if (solutiondelete) GlobalInfo->solutiondelete = solutiondelete;
  if (solutioncontextdelete)
    GlobalInfo->solutioncontextdelete = solutioncontextdelete;
  CHKMEMQ;

  PetscFunctionReturn(0);
}

/*! \page context Context handling
  
  In order to carry application-specific information and temporaries
  around, there are a few opaque handle contexts in syspro.

  A NumericalProblem structure is defined to have a context. This is
  cloned at the application of a preprocessor, and deleted when its
  application is finished.

  See SysProGetContextFunctions(), SysProProblemCloneContext(),
  SysProProblemDeleteContext().
 */
#undef __FUNCT__
#define __FUNCT__ "SysProGetContextFunctions"
static PetscErrorCode SysProGetContextFunctions
(PetscErrorCode(**ctxcloner)(const char*,const char*,void*,void**),
 PetscErrorCode(**ctxfree)(void*)
 )
{
  PetscFunctionBegin;
  if (ctxcloner) *ctxcloner = GlobalInfo->clonecontext;
  if (ctxfree) *ctxfree = GlobalInfo->freecontext;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

static PetscErrorCode SysProProblemCloneContext
(const char *preprocessor,const char *type,NumericalProblem in,NumericalProblem out)
{
  PetscErrorCode(*cloner)(const char*,const char*,void*,void**),ierr;
  PetscFunctionBegin;
  ierr = SysProGetContextFunctions(&cloner,PETSC_NULL); CHKERRQ(ierr);
  if (cloner) {
    ierr = (*cloner)(preprocessor,type,in->ctx,&(out->ctx)); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

static PetscErrorCode SysProProblemDeleteContext(NumericalProblem problem)
{
  PetscErrorCode(*freer)(void*),ierr;
  PetscFunctionBegin;
  ierr = SysProGetContextFunctions(PETSC_NULL,&freer); CHKERRQ(ierr);
  /*
  if (freer && problem->ctx) {
    ierr = (*freer)(problem->ctx); CHKERRQ(ierr);
  }
  */
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RegisterPreprocessorSetting"
static PetscErrorCode RegisterPreprocessorSetting
(const char *preprocess,const char *type,int option)
{
  PetscFunctionBegin;
  currentpreprocessors[preprocesslevel] = preprocess;
  currentchoices[preprocesslevel] = type;
  currentoptions[preprocesslevel] = option;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessorGetSetting"
PetscErrorCode PreprocessorGetSetting
(const char *preprocess,const char **type,int *option)
{
  int i;
  PetscFunctionBegin;
  for (i=0; i<=preprocesslevel; i++) {
    if (!strcmp(currentpreprocessors[i],preprocess)) {
      *type = currentchoices[i];
      *option = currentoptions[i];
      goto found;
    }
  }
  SETERRQ1(PETSC_COMM_SELF,1,"Prprocessor <%s> is not active",preprocess);
 found:
  CHKMEMQ;
  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "RetrievePreprocessorChoice"
PetscErrorCode RetrievePreprocessorChoice(int idx,const char **type,int *option)
{
  PetscFunctionBegin;
  *type = currentchoices[idx];
  *option = currentoptions[idx];
  CHKMEMQ;
  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "PreprocessorGetIndex"
PetscErrorCode PreprocessorGetIndex(const char *name,int *prenumber)
{
  int ipre; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  *prenumber = -1;
  for (ipre=0; ipre<npreprocess; ipre++) {
    ierr = PetscStrcmp(name,preprocessors[ipre]->name,&flg); CHKERRQ(ierr);
    if (flg) {
      *prenumber = ipre;
      goto found;
    }
  }
  SETERRQ1(PETSC_COMM_SELF,1,"Could not find preprocessor <%s>",name);
 found:
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SystemPreprocessorGetByName"
PetscErrorCode SystemPreprocessorGetByName
(const char *name,SystemPreprocessor *pp)
{
  int ipre; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PreprocessorGetIndex(name,&ipre); CHKERRQ(ierr);
  *pp = preprocessors[ipre];
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformGetByName"
PetscErrorCode TransformGetByName(const char *name,SalsaTransform *tf)
{
  SystemPreprocessor pp; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SystemPreprocessorGetByName(name,&pp); CHKERRQ(ierr);
  *tf = pp->transform;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

/*! \page context Preprocessor contexts
  In order to offer flexibility, there are several possibilities
  for user objects to be stored under opaque (\c void*) pointers.

  \section Global contexts

  Each preprocessor that has declared a \c contextcreate
  function, will create that context at the start of its traversal.
  This context is then globally registered with RegisterPreprocessorContext()
  under the name of this preprocessor, so that other preprocessor
  can have access to it with PreprocessorGetContext().

  There is one special context, which is stored under the \c solution
  handle. This one is not created by default. See the "linear" package
  for an example of how to use it.

  \section Local contexts
  
  The \c start_function can create a context, which is input
  to the \c end_function. This context serves to preser data that is
  necessary for the inverse transformation that is applied in the
  \c start_function
*/
#undef __FUNCT__
#define __FUNCT__ "RegisterPreprocessorContext"
PetscErrorCode RegisterPreprocessorContext(const char *pre,void *ctx)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (!preprocessorcontexts) {
    SETERRQ(PETSC_COMM_SELF,1,"Preprocessor contexts array should have been allocated");
  } else {
    PetscBool flg; int ipre;
    ierr = PetscStrcmp(pre,"solution",&flg); CHKERRQ(ierr);
    if (flg) {
      solutioncontext = ctx;
    } else {
      ierr = PreprocessorGetIndex(pre,&ipre); CHKERRQ(ierr);
      preprocessorcontexts[ipre] = ctx;
    }
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessorGetContext"
PetscErrorCode PreprocessorGetContext(const char *pre,void **ctx)
{
  PetscBool flg; int ipre; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscStrcmp(pre,"solution",&flg); CHKERRQ(ierr);
  if (flg) {
    *ctx = solutioncontext;
  } else {
    ierr = PreprocessorGetIndex(pre,&ipre); CHKERRQ(ierr);
    if (!preprocessorcontexts[ipre])
      SETERRQ1(PETSC_COMM_SELF,1,"No context set for preprocessor <%s>",pre);
    *ctx = preprocessorcontexts[ipre];
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

/*! Setup actions that are particular to a specific class of preprocessors,
  such as scalings. This performs the following actions:
  - any user specified setup, for instance disqualifying preprocessors
  on purely logistical grounds.
  - subsequently, the suitability functions (\ref suit) are evaluated for each
  preprocessor and the current problem. In the current implementation
  this is only used to disqualify preprocessors. We'll get more
  sophisticated later.
*/
#undef __FUNCT__
#define __FUNCT__ "PreprocessorSpecificSetup"
static PetscErrorCode PreprocessorSpecificSetup
(const char *preprocess,NumericalProblem problem,PetscBool user_choices)
{
  SystemPreprocessor pp; SalsaTransform tf; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SystemPreprocessorGetByName(preprocess,&pp); CHKERRQ(ierr);
  ierr = TransformGetByName(preprocess,&tf); CHKERRQ(ierr);

  if (GlobalInfo->problemmonitor) {
    ierr = (GlobalInfo->problemmonitor)(problem); CHKERRQ(ierr);
  }

  /* if the user specified what choices to take, all others have
     already been disabled
  */
  if (!user_choices) {
    int i;
    /* declare all preprocessor choices as suitable, then 
       eliminate some if there is a setup function for it
    */
    ierr = PreprocessorApplyAprioriSelection(pp); CHKERRQ(ierr);
    /*ierr = TransformObjectsUnmarkAll(tf); CHKERRQ(ierr);*/
    if (pp->setup) {
      ierr = (pp->setup)(problem,tf); CHKERRQ(ierr);
    }
    /* Now evaluate any suitability functions, and mark preprocessors 
       accordingly.
    */
    for (i=0; i<tf->n_objects; i++) {
      SalsaTransformObject to = tf->transformobjects[i]; void *sctx;
      PetscErrorCode(*f)(NumericalProblem,void*,SuitabilityValue*);
      ierr = TransformObjectGetSuitabilityFunction(to,&sctx,&f); CHKERRQ(ierr);
      if (f) {
	SuitabilityValue v; const char *name;
	ierr = TestSuitabilityContext(tf,sctx); CHKERRQ(ierr);
	ierr = (*f)(problem,sctx,&v); CHKERRQ(ierr);
	ierr = TransformObjectGetName(to,&name); CHKERRQ(ierr);
	printf("method %s compute suitability as %e\n",name,v);
	if (v==0) {
	  ierr = TransformObjectMark(to); CHKERRQ(ierr);
	}
      }
    }
  }
  ierr = ReportEnabledPreprocessors(preprocess); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysproPreprocessorStartFunction"
static PetscErrorCode SysproPreprocessorStartFunction
(SystemPreprocessor preprocessor,const char *type,int option,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *preprocessor_context,void **transform_context, PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (!(preprocessor->start_function)) 
    SETERRQ1(PETSC_COMM_SELF,1,"Preprocessor <%s> has no start function",preprocessor->name);
  ierr = (preprocessor->start_function)
    (type,option,overwrite,inproblem,outproblem,
     preprocessor_context,transform_context,success); CHKERRQ(ierr);
  if (!*success) PetscFunctionReturn(0);
  /* what on earth ?!
  if (GlobalInfo->classproblemcloner) {
    ierr = (GlobalInfo->classproblemcloner)
      (preprocessor->name,type,option,inproblem,*outproblem); CHKERRQ(ierr);
  } else *outproblem = inproblem;
  */
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProPreprocessorEndFunction"
static PetscErrorCode SysProPreprocessorEndFunction
(SystemPreprocessor preprocessor,const char *pclassname,const char *type, PetscBool do_not_keep_original,void *gctx,void *ctx,
 NumericalProblem next_problem,NumericalProblem problem, NumericalSolution next_solution,NumericalSolution *rsolution)
{
  NumericalSolution solution; PetscErrorCode ierr;
  PetscFunctionBegin;

  /* if there are separate solutions on this level and the surrounding,
   * the outer one is created here and the nested one is deleted.
   */
  if (next_solution && (preprocessor->end_function || GlobalInfo->solutioncopy)) {
    if (!GlobalInfo->solutioncreator) SETERRQ1(PETSC_COMM_SELF,1,"Preprocessor %s requires solution creator",pclassname);
    if (!GlobalInfo->solutiondelete)  SETERRQ1(PETSC_COMM_SELF,1,"Preprocessor %s requires solution destructor",pclassname);
    ierr = (GlobalInfo->solutioncreator)(problem,&solution); CHKERRQ(ierr);
    if (preprocessor->end_function) {
      ierr = (preprocessor->end_function)
	(type,do_not_keep_original,gctx,ctx,next_problem,problem,next_solution,solution); CHKERRQ(ierr);
    } else if (GlobalInfo->solutioncopy && next_solution) {
      ierr = (GlobalInfo->solutioncopy)(next_solution,solution); CHKERRQ(ierr);
    }
    ierr = (GlobalInfo->solutiondelete)(next_solution); CHKERRQ(ierr);
  } else solution = next_solution;

  if (next_problem!=problem) {
    if (!GlobalInfo->problemdelete)  SETERRQ1(PETSC_COMM_SELF,1,"Preprocessor %s requires problem destructor",pclassname);
    ierr = (GlobalInfo->problemdelete)(next_problem); CHKERRQ(ierr);
  }
  ierr = SysProProblemDeleteContext(next_problem); CHKERRQ(ierr);
  if (!do_not_keep_original) { /* deallocate next_problem */ }
  if (GlobalInfo->errortracer) {
    char caption[200];
    sprintf(caption,"After preprocessor <%s>=<%s>",pclassname,type);
    ierr = (GlobalInfo->errortracer)(problem,solution,caption); CHKERRQ(ierr);
  }
  *rsolution  = solution;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ChooseFirstTransform"
static PetscErrorCode ChooseFirstTransform
(NumericalProblem problem,const char *preprocess,
 PetscBool user_choices,PetscBool exhaustive,
 SalsaTransform transform,SystemPreprocessor preprocessor,
 SalsaTransformObject *transformitem,PetscBool *moretransform,
 const char **type)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (user_choices || exhaustive || !preprocessor->intelligence ) {
    ierr = TransformGetNextUnmarkedItem
      (transform,NULL,transformitem,moretransform); CHKERRQ(ierr);
    if (!*moretransform)
      SETERRQ1(PETSC_COMM_SELF,1,"Can not find any <%s>\n",preprocessor->name);
    ierr = TransformObjectGetName(*transformitem,type); CHKERRQ(ierr);
    ierr = SysProTraceMessage
      ("Preprocessor choice: %s=%s\n",preprocess,*type); CHKERRQ(ierr);
  } else {
    const char *reason;
    /* 
     * Intelligent setting:
     * Compute required categories
     */
    *moretransform = PETSC_TRUE;
    if (GlobalInfo->computecategory) {
      char *c;
      c = strtok((char*)preprocessor->required,",");
      while (c) {
	ierr = (GlobalInfo->computecategory)(c,problem); CHKERRQ(ierr);
	c = strtok(NULL,",");
      }
    }
    ierr = (preprocessor->intelligence)(problem,type,&reason); CHKERRQ(ierr);
    ierr = SysProTraceMessage
	("Intelligent choice <%s=%s> because <%s>\n",preprocess,*type,reason); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessedSolution"
/*!
  This routine handles the application of one preprocessor. Depending on
  the runtime setup (see section \ref modes), one choice is applied, or
  a sequence of choices is applied consecutively. The forward and backward
  transformation of the preprocessor are done here, and if necessary, backup
  copies of the system are kept around.
 */
PetscErrorCode PreprocessedSolution
(const char *pclassname,NumericalProblem problem,void *prevctx,NumericalSolution *rsolution)
{
  NumericalSolution solution=NULL; SystemPreprocessor preprocessor;
  SalsaTransform transform; SalsaTransformObject transformitem;
  PetscBool exhaustive,do_not_keep_original,success,moretransform,user_choices;
  const char *type;
  void *preprocessor_context;
  PetscErrorCode ierr;

  PetscFunctionBegin;

  ierr = ReportSysProCallStackState(pclassname); CHKERRQ(ierr);

  ierr = SystemPreprocessorGetByName(pclassname,&preprocessor); CHKERRQ(ierr);
  transform = preprocessor->transform; exhaustive = preprocessor->exhaustive;
  ierr = TransformGetUserChoices(transform,&user_choices); CHKERRQ(ierr);
  CHKMEMQ;

  /*
   * Do a global setup that is identical for all preprocessors,
   * for instance, compute problem characteristics
   */
  if (GlobalInfo->classdynamicsetup) {ierr = (GlobalInfo->classdynamicsetup)(pclassname,problem); CHKERRQ(ierr);}

  /*
   * Do setup based on the current matrix and preprocessor,
   * for instance, disable certain preprocessor choices.
   */
  ierr = PreprocessorSpecificSetup(pclassname,problem,user_choices); CHKERRQ(ierr);

  /*
   * Create a context object that will exist for the duration of
   * this preprocessor
   */
  if (preprocessor->ctxcreate) {
    ierr = (preprocessor->ctxcreate)(problem,&preprocessor_context); CHKERRQ(ierr);
    ierr = RegisterPreprocessorContext(pclassname,preprocessor_context); CHKERRQ(ierr);
  }

  /*
   * Pick the first or only choice
   */
  ierr = ChooseFirstTransform(problem,pclassname,user_choices,exhaustive,transform,preprocessor,
			      &transformitem,&moretransform,&type); CHKERRQ(ierr);

  /*
   * Do we need to back up the original system?
   */
  do_not_keep_original = TRUTH(!exhaustive && !GlobalInfo->errortracer);
  CHKMEMQ;

  /*
   * Loop over preprocessor choices
   */
  while (moretransform) {
    const char *next_solver; int option;
    PetscBool optionsloop,optionflg,solution_is_allocated = PETSC_FALSE;

    optionsloop = exhaustive;

    /*
     * Loop over option values; if there are no options, force one run through.
     */
    ierr = TransformItemGetFirstOption(pclassname,type,&option,&optionflg); CHKERRQ(ierr);
    if (!optionflg) {optionflg = PETSC_TRUE; option = 0;}

    while (optionflg) {
      void *transform_context; PetscBool didaniteration=PETSC_TRUE;
      NumericalProblem next_problem; NumericalSolution next_solution;

      ierr = SysProTraceMessage("Preprocessor <%s=%s,%d>\n",pclassname,type,option); CHKERRQ(ierr);
      ierr = RegisterPreprocessorSetting(pclassname,type,option); CHKERRQ(ierr);

      /*
       * Apply the preprocessor
       */
      solution_is_allocated = PETSC_FALSE;
      ierr = SysproPreprocessorStartFunction(preprocessor,type,option,do_not_keep_original,problem,&next_problem,
					     preprocessor_context,&transform_context,&success); CHKERRQ(ierr);
      if (!success) {
	ierr = SysProTraceMessage("Preprocessor breakdown <%s=%s,%d>\n",pclassname,type,option); CHKERRQ(ierr);
	didaniteration = PETSC_FALSE; goto next;
      }
      ierr = SysProProblemCloneContext(pclassname,type,problem,next_problem); CHKERRQ(ierr);
      CHKMEMQ;

      /* 
       * Here is where we recursively call the next preprocessor or the final problem solver
       */
      ierr = SuccessorPreprocessor(pclassname,&next_solver); CHKERRQ(ierr);
      preprocesslevel++;
      if (next_solver) {
	ierr = PreprocessedSolution(next_solver,next_problem,transform_context,&next_solution); CHKERRQ(ierr);
      } else if (GlobalInfo->problemsolver) {
	ierr = (GlobalInfo->problemsolver)(next_problem,transform_context,&next_solution); CHKERRQ(ierr);
      } else SETERRQ(PETSC_COMM_SELF,1,"No problem solving routine set");
      preprocesslevel--;
      CHKMEMQ;

      /*
       * Un-preprocess, and delete inner solution
       */
      ierr = SysProPreprocessorEndFunction
	(preprocessor,pclassname,type,do_not_keep_original,preprocessor_context,transform_context,
	 next_problem,problem,next_solution,&solution); CHKERRQ(ierr);
      solution_is_allocated = (PetscBool)(solution!=next_solution);
    next:
      CHKMEMQ;
      optionflg = optionsloop;
      if (optionflg) {ierr = TransformItemGetNextOption(pclassname,type,&option,&optionflg); CHKERRQ(ierr);}
      if (didaniteration && optionflg && solution) {
	/* if we do another iteration, delete the solution we just created;
	   the null solution case hapens if the inner solve crashed.
	*/
	if (GlobalInfo->solutiondelete) {
	  ierr = (GlobalInfo->solutiondelete)(solution); CHKERRQ(ierr);
	} else
	  SETERRQ1(PETSC_COMM_SELF,1,"Option looping (for %s) requires solution destructor",pclassname);
	if (GlobalInfo->solutioncontextdelete) {
	  /* ierr = (GlobalInfo->solutioncontextdelete)(next_solution); CHKERRQ(ierr); */
	}
      }
    } /* end of options loops */
    CHKMEMQ;

    if (exhaustive) {
      ierr = TransformGetNextUnmarkedItem(transform,type,&transformitem,&moretransform); CHKERRQ(ierr);
      if (moretransform) {ierr = TransformObjectGetName(transformitem,&type); CHKERRQ(ierr);}
    } else moretransform = PETSC_FALSE;
    if (solution && moretransform) {
      /* if we do another iteration, delete temporary stuff; solution can not exist if the preprocessor broke down */
      if (GlobalInfo->solutioncontextdelete) {ierr = (GlobalInfo->solutioncontextdelete)(solution); CHKERRQ(ierr);}
      if (GlobalInfo->solutiondelete) {ierr = (GlobalInfo->solutiondelete)(solution); CHKERRQ(ierr);
      } else
	SETERRQ1(PETSC_COMM_SELF,1,"Preprocessor looping (for %s) requires solution destructor",preprocessor);
    }
  } /* end of preprocessor loop */

  if (preprocessor->unset) {ierr = (preprocessor->unset)(problem); CHKERRQ(ierr);}
  if (preprocessor->ctxdelete) {ierr = (preprocessor->ctxdelete)(preprocessor_context); CHKERRQ(ierr);}
  CHKMEMQ;

  *rsolution = solution;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessedProblemSolving"
/*! Invoking this routine starts the preprocessing and ultimate solution
  of the numerical problem.
*/
PetscErrorCode PreprocessedProblemSolving
(NumericalProblem problem,NumericalSolution *solution)
{
  const char *preprocess; PetscErrorCode ierr;
  PetscFunctionBegin;
  preprocesslevel = 0;
  ierr = GetFirstPreprocessor(&preprocess); CHKERRQ(ierr);
  if (!preprocess) {
    if (GlobalInfo->problemsolver) {
      ierr = (GlobalInfo->problemsolver)(problem,NULL,solution); CHKERRQ(ierr);
      if (GlobalInfo->errortracer) {
	ierr = (GlobalInfo->errortracer)
	  (problem,*solution,"After main solve"); CHKERRQ(ierr);
      }
    } else SETERRQ(PETSC_COMM_SELF,1,"No problem solving routine set");
  } else {
    ierr = PreprocessedSolution
      (preprocess,problem,NULL,solution); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}
