#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "syspro.h"
#include "sysprotransform.h"
#include "syspro_impl.h"

/*! \page reporting Preprocessor reporting
  For purposes of reporting, there are routines for retrieving the names
  of all preprocessors in sequence. This is the general idea:
\verbatim
  ierr = GetFirstPreprocessor(&name); CHKERRQ(ierr);
  while (name) {
   .....
   ierr = GetNextPreprocessor(&name); CHKERRQ(ierr);
  }
\endverbatim
  Similarly, StartRetrievingCurrentPreprocessors() and
  ContinueRetrievingCurrentPreprocessors() get the class and specific
  choice of the currently active preprocessors.

  With StartRetrievingAllPreprocessors() and
  ContinueRetrievingAllPreprocessors() one can the classes, and
  in each class all defined choices.

  The specific reporting funtions are:
  - TabReportPreprocessors() : report on currently active preprocessors; 
  suitable for database file output
  - ReportEnabledPreprocessors() : report on non-disabled choices for
  a given preprocessor
  - ReportSysProCallStackState() : report currently active preprocessors
*/
extern int npreprocess,preprocesslevel;
static int preprocessreadout;
extern SystemPreprocessor *preprocessors;

#undef __FUNCT__
#define __FUNCT__ "GetFirstPreprocessor"
/*!
  Get the name of the first declared preprocessor (in order of declaration)
  or null if none have been declared.
  Subsequent preprocessors can be retrieved with
  GetNextPreprocessor() or SuccessorPreprocessor().
*/
PetscErrorCode GetFirstPreprocessor(const char **preprocess)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  preprocessreadout = 0;
  ierr = GetNextPreprocessor(preprocess); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetNextPreprocessor"
/*! 
  Get the next preprocessor according to the variable preprocessreadout.

  The result is null if there are no further preprocessors.
*/
PetscErrorCode GetNextPreprocessor(const char **next_one)
{
  PetscFunctionBegin;
  if (preprocessreadout>=npreprocess)
    *next_one = NULL;
  else
    *next_one = preprocessors[preprocessreadout++]->name;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SuccessorPreprocessor"
/*! Given a preprocessor, get the name of the next one
  (in order of declaration) or null if there are no further ones.

  The arguments are allowed to be the same.
*/
PetscErrorCode SuccessorPreprocessor(const char *this_one,const char **next_one)
{
  int ipre; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  for (ipre=0; ipre<npreprocess; ipre++) {
    ierr = PetscStrcmp(this_one,preprocessors[ipre]->name,&flg); CHKERRQ(ierr);
    if (flg) {
      if (ipre<npreprocess-1)
	*next_one = preprocessors[ipre+1]->name;
      else
	*next_one = 0;
      goto found;
    }
  }
  SETERRQ1(PETSC_COMM_SELF,1,"Could not find preprocessor <%s>",this_one);
 found:
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StartRetrievingCurrentPreprocessors"
/*!
  This routine gives the class and current value of the first
  declared preprocessor.
  To get the next preprocessor, call 
  ContinueRetrievingAllPreprocessors().

  The class, types, and ntypes arguments can all be null.
*/
PetscErrorCode StartRetrievingCurrentPreprocessors
(const char **cclass,const char **type,int *opt,PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  preprocessreadout = 0;
  ierr = ContinueRetrievingCurrentPreprocessors
    (cclass,type,opt,success); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ContinueRetrievingCurrentPreprocessors"
/*!
  This routine is to be used repeatedly after an initial call
  to StartRetrievingCurrentPreprocessors().
*/
PetscErrorCode ContinueRetrievingCurrentPreprocessors
(const char **cclass,const char **atype,int *opt,PetscBool *success)
{
  PetscFunctionBegin;
  if (preprocessreadout>=preprocesslevel) 
    *success = PETSC_FALSE;
  else {
    const char *curtype; int curopt; PetscErrorCode ierr;
    ierr = RetrievePreprocessorChoice
      (preprocessreadout,&curtype,&curopt); CHKERRQ(ierr);
    *success = PETSC_TRUE;
    if (cclass) *cclass = preprocessors[preprocessreadout]->name;
    if (atype)  *atype = curtype;
    if (opt)   *opt = curopt;
    preprocessreadout++;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "StartRetrievingAllPreprocessors"
/*!
  This routine gives the class of the first
  declared preprocessor, and all possible values.
  To get the next preprocessor, call 
  ContinueRetrievingAllPreprocessors().

  The class, types, and ntypes arguments can all be null.

  The \c types argument is allocated internally and should be deallocated
  by the user.
*/
PetscErrorCode StartRetrievingAllPreprocessors
(const char **cclass,const char ***types,int *ntypes,PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  preprocessreadout = 0;
  ierr = ContinueRetrievingAllPreprocessors
    (cclass,types,ntypes,success); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "InitRetrievingPreprocessors"
PetscErrorCode InitRetrievingPreprocessors()
{
  PetscFunctionBegin;
  preprocessreadout = 0;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ContinueRetrievingAllPreprocessors"
/*!
  This routine is to be used repeatedly after an initial call
  to StartRetrievingAllPreprocessors().
  
  The \c types argument is allocated internally and should be deallocated
  by the user.
*/
PetscErrorCode ContinueRetrievingAllPreprocessors
(const char **cclass,const char ***types,int *ntypes,PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (preprocessreadout>=npreprocess) 
    *success = PETSC_FALSE;
  else {
    *success = PETSC_TRUE;
    if (cclass)
      *cclass = preprocessors[preprocessreadout]->name;
    if (ntypes)
      *ntypes = preprocessors[preprocessreadout]->transform->n_objects;
    if (types) {
      ierr = TransformObjectsGetNames
	(preprocessors[preprocessreadout]->transform,types); CHKERRQ(ierr);
    }
    preprocessreadout++;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "RetrieveAllPreprocessorValues"
PetscErrorCode RetrieveAllPreprocessorValues
(const char *cclass,const char ***types,int *ntypes)
{
  int idx; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  for (idx=0; idx<npreprocess; idx++) {
    ierr = PetscStrcmp(cclass,preprocessors[idx]->name,&flg); CHKERRQ(ierr);
    if (flg) {preprocessreadout=idx; goto retrieve;}
  }
  SETERRQ1(PETSC_COMM_SELF,1,"Preprocessor unknown: <%s>",cclass);
 retrieve:
  ierr = ContinueRetrievingAllPreprocessors
    (PETSC_NULL,types,ntypes,&flg); CHKERRQ(ierr);
  if (!flg) SETERRQ(PETSC_COMM_SELF,1,"This can not happen");
  PetscFunctionReturn(0);
}

#define LINELEN 1500
#define REPOSITION(a,b) \
  ierr = PetscStrlen(a,&b); CHKERRQ(ierr); \
  if (b>LINELEN) SETERRQ(PETSC_COMM_SELF,1,"string overflow")
#undef __FUNCT__
#define __FUNCT__ "TabReportPreprocessors"
/*! Report all defined preprocessors. Either key and val argument can 
   be NULL. The string arguments returned need to be deallocated
   in the calling environment.
*/
static PetscErrorCode TabReportPreprocessors
(PetscBool active,const char **key,const char **val,int separator)
{
  PetscBool has_preprocessor; const char *cclass,*value,*sql_key,*sql_val;
  size_t loc_key,loc_val; int option; 
  PetscBool begun=PETSC_FALSE; PetscErrorCode ierr;

  PetscFunctionBegin;
  if (key) {
    ierr = PetscMalloc(LINELEN*sizeof(char),&sql_key); CHKERRQ(ierr);
    ierr = PetscMemzero((char*)sql_key,LINELEN*sizeof(char)); CHKERRQ(ierr);
  }
  if (val) {
    ierr = PetscMalloc(LINELEN*sizeof(char),&sql_val); CHKERRQ(ierr);
    ierr = PetscMemzero((char*)sql_val,LINELEN*sizeof(char)); CHKERRQ(ierr);
  }
  loc_key = loc_val = 0;

  ierr = InitRetrievingPreprocessors(); CHKERRQ(ierr);
  do {
    if (active) {
      ierr = ContinueRetrievingCurrentPreprocessors
	(&cclass,&value,&option,&has_preprocessor); CHKERRQ(ierr);
      if (has_preprocessor) {
      } else active = PETSC_FALSE;
    }
    if (!active) {
      ierr = ContinueRetrievingAllPreprocessors
	(&cclass,PETSC_NULL,PETSC_NULL,&has_preprocessor); CHKERRQ(ierr);
      value = "(inactive)"; option = 0;
      if (has_preprocessor) {
      } else break;      
    }

    if (begun) {
      if (key) sprintf((char*)sql_key+loc_key++,"\t");
      if (val) sprintf((char*)sql_val+loc_val++,"\t");
    } else begun=PETSC_TRUE;

    if (key) sprintf((char*)sql_key+loc_key,"%s\t%sv",cclass,cclass);
    if (val) sprintf((char*)sql_val+loc_val,"%s\t%d",value,option);

    if (key) REPOSITION((char*)sql_key,loc_key);
    if (val) REPOSITION((char*)sql_val,loc_val);

  } while (1);
  CHKMEMQ;

  if (key) *key = sql_key;
  if (val) *val = sql_val;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TabReportAllPreprocessors"
/* Report all defined preprocessors. Either key and val argument can 
   be NULL.
*/
PetscErrorCode TabReportAllPreprocessors(const char **key,int separator)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TabReportPreprocessors
    (PETSC_FALSE,key,PETSC_NULL,separator); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TabReportActivePreprocessors"
/* Report all defined preprocessors. Either key and val argument can 
   be NULL.
*/
PetscErrorCode TabReportActivePreprocessors(const char **key,const char **val,int separator)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TabReportPreprocessors(PETSC_TRUE,key,val,separator); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "ScreenOutputTab"
PetscErrorCode ScreenOutputTab(const char *key,const char *val)
{
  PetscBool flg;
  char *ck,*cv; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = SysProHasTrace(&flg); CHKERRQ(ierr);
  if (flg) {
    ck = (char*)key; cv = (char*)val;
    while (1) {
      size_t ik=0, iv=0, lk,lv;
      lk = strlen(ck); while (ik<lk && ck[ik]!='\t') ik++; if (ik<lk) ck[ik] = 0;
      lv = strlen(cv); while (iv<lv && cv[iv]!='\t') iv++; if (iv<lv) cv[iv] = 0;
      ierr = SysProTraceMessage("%s: %s\n",ck,cv); CHKERRQ(ierr);
      if (iv<lv) {
	ck[ik] = '\t'; cv[iv] = '\t';
	ck = ck+ik+1; cv = cv+iv+1;
      } else break;
    }
  }
  CHKMEMQ;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ScreenOutputTabLine"
PetscErrorCode ScreenOutputTabLine(const char *key,const char *val)
{
  PetscBool flg;
  char *ck,*cv,*message; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = SysProHasTrace(&flg); CHKERRQ(ierr);
  if (flg) {
#define MAXLEN 500
    ierr = PetscMalloc(MAXLEN*sizeof(char),&message); CHKERRQ(ierr);
    ierr = PetscMemzero(message,MAXLEN*sizeof(char)); CHKERRQ(ierr);
    ck = (char*)key; cv = (char*)val;
    while (1) {
      int l = strlen(message);
      size_t ik=0, iv=0, lk,lv;

      lk = strlen(ck);
      while (ik<lk && ck[ik]!='\t') ik++;
      if (ik<lk) ck[ik] = 0;

      lv = strlen(cv);
      while (iv<lv && cv[iv]!='\t') iv++;
      if (iv<lv) cv[iv] = 0;

      if (l+strlen(ck)+strlen(cv)+6<MAXLEN) {
	sprintf((char*)message+l,"%s: %s || ",ck,cv);
      }
      if (iv<lv) {
	ck[ik] = '\t'; cv[iv] = '\t';
	ck = ck+ik+1; cv = cv+iv+1;
      } else break;
    }
    ierr = SysProTraceMessage("%s\n",message); CHKERRQ(ierr);
    ierr = PetscFree(message); CHKERRQ(ierr);
  }
  CHKMEMQ;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ReportEnabledPreprocessors"
/*! Report preprocessor choices that are available after the 
  specific setup has possible disabled some of the registered ones.
  This function uses the \c sysprotrace function, so this has to have
  been declared.
*/
PetscErrorCode ReportEnabledPreprocessors(const char *name)
{
  PetscBool flg;
  SalsaTransform tf; const char *s; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SysProHasTrace(&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = TransformGetByName(name,&tf); CHKERRQ(ierr);
    ierr = TransformReportEnabled(tf,&s); CHKERRQ(ierr);
    ierr = SysProTraceMessage
      ("Enabled <%s> preprocessors: %s\n",name,s); CHKERRQ(ierr);
    ierr = PetscFree(s); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ReportSysProCallStackState"
/*! Report preprocessor choices that are available after the 
  specific setup has possible disabled some of the registered ones.
  This function uses the \c sysprotrace function, so this has to have
  been declared.
*/
PetscErrorCode ReportSysProCallStackState(const char *name)
{
  PetscBool flg;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SysProHasTrace(&flg); CHKERRQ(ierr);
  if (flg) {
    const char *key,*val;
    ierr = SysProTraceMessage
      ("Entering preprocessor: <%s>\n",name); CHKERRQ(ierr);
    ierr = SysProTraceMessage("Currently active:\n"); CHKERRQ(ierr);
    ierr = TabReportPreprocessors(PETSC_TRUE,&key,&val,','); CHKERRQ(ierr);
    ierr = ScreenOutputTabLine(key,val); CHKERRQ(ierr);
    ierr = PetscFree(key); CHKERRQ(ierr);
    ierr = PetscFree(val); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}
