#include <stdlib.h>
#include "anamod.h"
#include "syspro.h"
#include "sysprolinear.h"

/*! \ref anamod SysPro-AnaMod interface

  The SysPro linear package has a few routines to facilitate integration
  with AnaMod

  - \c SysProComputeQuantity() : to compute a quantity using AnaMod
  and store it as the metadata of a linear system
  - \c SysProRetrieveQuantity() : to get an already computed quantity
  - \c SysProFreeQuantities() : to destroy the metadata object
  - \c SysProRemoveQuantity() : to invalidate/free selected quantities
*/
#undef __FUNCT__
#define __FUNCT__ "SysProComputeQuantity"
/*! This routine is used in SysPro to compute quantities.
  See also SysProRetrieveQuantity().
*/
PetscErrorCode SysProComputeQuantity
(NumericalProblem theproblem,const char *cat,const char *cmp,
 void *res,int *reslen,PetscBool *flg)
{
  LinearSystem problem = (LinearSystem)theproblem;
  Mat A;
  NMD_metadata nmd;
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = LinearSystemGetParts
    (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  ierr = ComputeQuantity((void*)A,cat,cmp,(AnalysisItem*)res,reslen,flg); CHKERRQ(ierr);
  if (*flg) {
    AnalysisDataType type;
    ierr = LinearSystemGetMetadata(problem,&nmd); CHKERRQ(ierr);
    ierr = GetDataType(cat,cmp,&type,NULL); CHKERRQ(ierr);
    ierr = NMDSetValue(nmd,cat,cmp,type,res); CHKERRQ(ierr);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProRetrieveQuantity"
/*! This routine is used in SysPro to retrieve already computed
  quantities. Reports failure if the quantity has not already been computed.
  See also SysProComputeQuantity().
*/
PetscErrorCode SysProRetrieveQuantity
(NumericalProblem theproblem,const char *cat,const char *cmp,
 void *res,int *reslen,PetscBool *flg)
{
  LinearSystem problem = (LinearSystem)theproblem;
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = HasComputeModule(cat,cmp,flg); CHKERRQ(ierr);
  if (*flg) {
    Mat A;
    ierr = LinearSystemGetParts
      (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
    ierr = HasQuantity((void*)A,cat,cmp,flg); CHKERRQ(ierr);
    if (*flg) {
      ierr = ComputeQuantity((void*)A,cat,cmp,(AnalysisItem*)res,reslen,flg); CHKERRQ(ierr);
    }
  }
  CHKMEMQ;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProRemoveQuantity"
/*! This routine is used to invalidate and free computed quantities.
  See also SysProRetrieveQuantity(), SysProComputeQuantity().
*/
PetscErrorCode SysProRemoveQuantity
(NumericalProblem theproblem,const char *cat,const char *cmp,PetscBool *flg)
{
  LinearSystem problem = (LinearSystem)theproblem;
  NMD_metadata nmd;
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = LinearSystemGetMetadata(problem,&nmd); CHKERRQ(ierr);
  ierr = NMDUnsetValue(nmd,cat,cmp); CHKERRQ(ierr);

  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProFreeQuantities"
PetscErrorCode SysProFreeQuantities(NumericalProblem theproblem)
{
  LinearSystem problem = (LinearSystem)theproblem;
  NMD_metadata nmd;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSystemGetMetadata(problem,&nmd); CHKERRQ(ierr);
  ierr = NMDDestroyObject(nmd); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
